"""Models for concrete occurrences of specific errors."""
from logging import getLogger

from datetime import datetime
from textwrap import dedent
from typing import Any, Dict, Mapping, List, Type, Union, Optional, Sequence
from pydantic import BaseModel, Field
from opentrons_shared_data.errors.codes import ErrorCodes
from .exceptions import ProtocolEngineError
from opentrons_shared_data.errors.exceptions import EnumeratedError

log = getLogger(__name__)


class ErrorOccurrence(BaseModel):
    """An occurrence of a specific error during protocol execution."""

    @classmethod
    def from_failed(
        cls: Type["ErrorOccurrence"],
        id: str,
        createdAt: datetime,
        error: Union[ProtocolEngineError, EnumeratedError],
    ) -> "ErrorOccurrence":
        """Build an ErrorOccurrence from the details available from a FailedAction or FinishAction."""
        if isinstance(error, ProtocolCommandFailedError) and error.original_error:
            wrappedErrors = [error.original_error]
        else:
            wrappedErrors = [
                cls.from_failed(id, createdAt, err) for err in error.wrapping
            ]
        return cls.construct(
            id=id,
            createdAt=createdAt,
            errorType=type(error).__name__,
            detail=error.message or str(error),
            errorInfo=error.detail,
            errorCode=error.code.value.code,
            wrappedErrors=wrappedErrors,
        )

    id: str = Field(..., description="Unique identifier of this error occurrence.")
    createdAt: datetime = Field(..., description="When the error occurred.")

    # Our Python should probably always set this to False--if we want it to be True,
    # we should probably be using a more specific subclass of ErrorOccurrence anyway.
    # However, we can't make this Literal[False], because we want this class to be able
    # to act as a catch-all for parsing defined errors that might be missing some
    # `errorInfo` fields because they were serialized by older software.
    isDefined: bool = Field(
        # default=False for database backwards compatibility, so we can parse objects
        # serialized before isDefined existed.
        default=False,
        description=dedent(
            """\
            Whether this error is *defined.*

            *Defined* errors have a strictly defined cause and effect, and we generally
            design them to be recoverable. For example, a `pickUpTip` command might
            return a `tipPhysicallyMissing` error, which is defined, and which you can
            recover from by doing a new `pickUpTip` in a different location.

            The `errorType` and `errorInfo` fields are useful for reacting to defined
            errors.

            *Undefined* errors are everything else. They could represent more obscure
            hardware failures that we aren't handling yet, or just bugs in our software.
            You can continue issuing commands to a run after it encounters an undefined
            error, but the robot may not behave well. For example, if a movement command
            fails with an undefined error, it could leave the robot in a messed-up
            state, and the next movement command could have bad path planning that
            collides the pipettes with stuff on the deck. We allow you to take this risk
            because it's sometimes better than safely stopping the run, which can
            waste time and reagents.
            """
        ),
    )

    errorType: str = Field(
        ...,
        description=dedent(
            """\
            This field has two meanings.

            When `isDefined` is `true`, this is a machine- and developer-readable
            identifier for what kind of error this is. You can use this to implement
            interactive error recovery flows, like watching for `"overpressure"` errors
            so you can tell the operator to unclog the tip.

            When `isDefined` is `false`, this is deprecated. For historical reasons,
            it will be the Python class name of some internal exception,
            like `"KeyError"` or `"UnexpectedProtocolError"`. The string values are not
            stable across software versions. If you're looking for something to show to
            robot operators, use `errorCode` and `detail` instead.
            """
        ),
    )

    errorCode: str = Field(
        default=ErrorCodes.GENERAL_ERROR.value.code,
        description=dedent(
            """\
            An enumerated error code for the error type.
            This is intended to be shown to the robot operator to direct them to the
            correct rough area for troubleshooting.
            """
        ),
    )

    detail: str = Field(
        ...,
        description=dedent(
            """\
            A short human-readable message about the error.

            This is intended to provide the robot operator with more specific details than
            `errorCode` alone. It should be no longer than a couple of sentences,
            and it should not contain internal newlines or indentation.

            It should not internally repeat `errorCode`, but it may internally repeat `errorType`
            if it helps the message make sense when it's displayed in its own separate block.
            """
        ),
    )

    errorInfo: Mapping[str, object] = Field(
        default={},
        description=dedent(
            """\
            Specific structured details about the error that may be useful for
            determining what happened.

            This might contain the same information as `detail` in a more structured form.
            It might also contain additional information that was too verbose or technical
            to put in `detail`.

            If `isDefined` is `true`, this object may have guaranteed contents, depending
            on `errorType`. If `isDefined` is `false`, you should not rely on any
            particular contents.
            """
        ),
    )

    wrappedErrors: List["ErrorOccurrence"] = Field(
        default=[], description="Errors that may have caused this one."
    )

    class Config:
        """Customize configuration for this model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: object) -> None:
            """Append the schema to make the errorCode appear required.

            `errorCode`, `wrappedErrors`, and `errorInfo` have defaults because they are not included in earlier
            versions of this model, _and_ this model is loaded directly from
            the on-robot store. That means that, without a default, it will
            fail to parse. Once a default is defined, the automated schema will
            mark this as a non-required field, which is misleading as this is
            a response from the server to the client and it will always have an
            errorCode defined. This hack is required because it informs the client
            that it does not, in fact, have to account for a missing errorCode, wrappedError, or errorInfo.
            """
            schema["required"].extend(["errorCode", "wrappedErrors", "errorInfo"])


# TODO (tz, 7-12-23): move this to exceptions.py when we stop relaying on ErrorOccurrence.
class ProtocolCommandFailedError(ProtocolEngineError):
    """Raised if a fatal command execution error has occurred."""

    def __init__(
        self,
        original_error: Optional[ErrorOccurrence] = None,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ProtocolCommandFailedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)
        self.original_error = original_error


ErrorOccurrence.update_forward_refs()



"""Script for starting up a python smoothie emulator."""
import logging
import asyncio

from opentrons.hardware_control.emulation.smoothie import SmoothieEmulator
from opentrons.hardware_control.emulation.parser import Parser


from opentrons.hardware_control.emulation.run_emulator import run_emulator_server
from opentrons.hardware_control.emulation.settings import Settings


async def run(settings: Settings) -> None:
    """Run the smoothie emulator.

    Args:
        settings: emulator settings

    Returns:
        None
    """
    smoothie = SmoothieEmulator(parser=Parser(), settings=settings.smoothie)
    await run_emulator_server(
        host=settings.smoothie.host,
        port=settings.smoothie.port,
        emulator=smoothie,
    )


def main() -> None:
    """Entry point."""
    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)
    asyncio.run(run(Settings()))


if __name__ == "__main__":
    main()



"""Command models to start heating a Heater-Shaker Module."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


SetTargetTemperatureCommandType = Literal["heaterShaker/setTargetTemperature"]


class SetTargetTemperatureParams(BaseModel):
    """Input parameters to set a Heater-Shaker's target temperature."""

    moduleId: str = Field(..., description="Unique ID of the Heater-Shaker Module.")
    celsius: float = Field(..., description="Target temperature in °C.")


class SetTargetTemperatureResult(BaseModel):
    """Result data from setting a Heater-Shaker's target temperature."""


class SetTargetTemperatureImpl(
    AbstractCommandImpl[
        SetTargetTemperatureParams, SuccessData[SetTargetTemperatureResult]
    ]
):
    """Execution implementation of a Heater-Shaker's set temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self,
        params: SetTargetTemperatureParams,
    ) -> SuccessData[SetTargetTemperatureResult]:
        """Set a Heater-Shaker's target temperature."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        hs_module_substate = self._state_view.modules.get_heater_shaker_module_substate(
            module_id=params.moduleId
        )

        # Verify temperature from hs module view
        validated_temp = hs_module_substate.validate_target_temperature(params.celsius)

        # Allow propagation of ModuleNotAttachedError.
        hs_hardware_module = self._equipment.get_module_hardware_api(
            hs_module_substate.module_id
        )

        if hs_hardware_module is not None:
            await hs_hardware_module.start_set_temperature(validated_temp)

        return SuccessData(
            public=SetTargetTemperatureResult(),
        )


class SetTargetTemperature(
    BaseCommand[SetTargetTemperatureParams, SetTargetTemperatureResult, ErrorOccurrence]
):
    """A command to set a Heater-Shaker's target temperature."""

    commandType: SetTargetTemperatureCommandType = "heaterShaker/setTargetTemperature"
    params: SetTargetTemperatureParams
    result: Optional[SetTargetTemperatureResult]

    _ImplementationCls: Type[SetTargetTemperatureImpl] = SetTargetTemperatureImpl


class SetTargetTemperatureCreate(BaseCommandCreate[SetTargetTemperatureParams]):
    """A request to create a Heater-Shaker's set temperature command."""

    commandType: SetTargetTemperatureCommandType = "heaterShaker/setTargetTemperature"
    params: SetTargetTemperatureParams

    _CommandCls: Type[SetTargetTemperature] = SetTargetTemperature



"""Thermocycler plate lift handling for labware movement using gripper."""
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, AsyncGenerator, Optional
from opentrons.hardware_control.modules.thermocycler import Thermocycler
from opentrons.protocol_engine.types import LabwareLocation, ModuleLocation, ModuleModel
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.state.module_substates import ThermocyclerModuleId
from contextlib import asynccontextmanager

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler


class ThermocyclerPlateLifter:
    """Implementation logic for TC Gen2 plate lifting."""

    _state_store: StateStore

    def __init__(
        self,
        state_store: StateStore,
        equipment: EquipmentHandler,
        movement: MovementHandler,
    ) -> None:
        """Initialize a ThermocyclerPlateLifter instance."""
        self._state_store = state_store
        self._equipment = equipment
        self._movement = movement

    def _get_tc_hardware(
        self, labware_location: LabwareLocation
    ) -> Optional[Thermocycler]:
        if isinstance(labware_location, ModuleLocation):
            module_id = labware_location.moduleId
            if (
                self._state_store.modules.get_connected_model(module_id)
                == ModuleModel.THERMOCYCLER_MODULE_V2
            ):
                # We already verify that TC lid is open before moving labware. So,
                # only assert that that is still true.
                assert self._state_store.modules.get_thermocycler_module_substate(
                    module_id
                ).is_lid_open, (
                    "Thermocycler lid needs to be open before performing a plate lift"
                )
                return self._equipment.get_module_hardware_api(
                    ThermocyclerModuleId(labware_location.moduleId)
                )
        return None

    @asynccontextmanager
    async def lift_plate_for_labware_movement(
        self,
        labware_location: LabwareLocation,
    ) -> AsyncGenerator[None, None]:
        """Lift plate if moving labware from Thermocycler Gen2.

        The 'lift_plate' function unsticks the plate, but does not leave the plate lifted.
        In order to leave the plate in a "lifted" position when picking up with the gripper,
        we need to use another function of the TC that will move the lid motor without returning
        it until the plate has been moved.

        If we are moving labware from TC Gen2, then there is a chance that the labware
        will need unsticking so it can be picked up. TC gen1 is neither compatible with
        the gripper/OT3 nor does it have any plate lifting mechanism so we no-op for it.
        """
        thermocycler_hardware = self._get_tc_hardware(labware_location=labware_location)
        if thermocycler_hardware is not None:
            await self._movement.home(axes=None)
            await thermocycler_hardware.lift_plate()
            try:
                await thermocycler_hardware.raise_plate()
                yield
            except asyncio.CancelledError as e:
                raise e
            else:
                await thermocycler_hardware.return_from_raise_plate()
        else:
            yield



from typing import NamedTuple, Union, Dict, Optional

from opentrons.types import Mount, DeckLocation, Point
from opentrons.legacy_commands import publisher
from opentrons.hardware_control import SyncHardwareAPI, types as hw_types

from ._types import OffDeckType
from .core.common import ProtocolCore


class HardwareManager(NamedTuple):
    """Back. compat. wrapper for a removed class called `HardwareManager`.

    This interface will not be present in PAPIv3.
    """

    hardware: SyncHardwareAPI


class RobotContext(publisher.CommandPublisher):
    """
    A context for the movement system of the robot.

    The RobotContext class provides the objects, attributes, and methods that allow
    you to control robot motor axes individually.

    Its methods can command the robot to perform an action, like moving to an absolute position,
    controlling the gripper jaw, or moving individual pipette motors.

    Objects in this class should not be instantiated directly. Instead, instances are
    returned by :py:meth:`ProtocolContext.robot`.

    .. versionadded:: 2.20

    """

    def __init__(self, core: ProtocolCore) -> None:
        self._hardware = HardwareManager(hardware=core.get_hardware())

    @property
    def hardware(self) -> HardwareManager:
        return self._hardware

    def move_to(
        self,
        mount: Union[Mount, str],
        destination: Point,
        velocity: float,
    ) -> None:
        raise NotImplementedError()

    def move_axes_to(
        self,
        abs_axis_map: Dict[hw_types.Axis, hw_types.AxisMapValue],
        velocity: float,
        critical_point: Optional[hw_types.CriticalPoint],
    ) -> None:
        raise NotImplementedError()

    def move_axes_relative(
        self, rel_axis_map: Dict[hw_types.Axis, hw_types.AxisMapValue], velocity: float
    ) -> None:
        raise NotImplementedError()

    def close_gripper_jaw(self, force: float) -> None:
        raise NotImplementedError()

    def open_gripper_jaw(self) -> None:
        raise NotImplementedError()

    def axis_coordinates_for(
        self, mount: Union[Mount, str], location: Union[DeckLocation, OffDeckType]
    ) -> None:
        raise NotImplementedError()

    def plunger_coordinates_for_volume(
        self, mount: Union[Mount, str], volume: float
    ) -> None:
        raise NotImplementedError()

    def plunger_coordinates_for_named_position(
        self, mount: Union[Mount, str], position_name: str
    ) -> None:
        raise NotImplementedError()

    def build_axis_map(
        self, axis_map: Dict[hw_types.Axis, hw_types.AxisMapValue]
    ) -> None:
        raise NotImplementedError()



import sys
import argparse
import datetime
import enum

sys.path.append("/opt/opentrons-robot-server")

from opentrons.hardware_control.thread_manager import ThreadManager  # noqa: E402
from opentrons.hardware_control.ot3api import OT3API  # noqa: E402
from opentrons.hardware_control.types import Axis, OT3Mount  # noqa: E402
from opentrons.hardware_control import HardwareControlAPI  # noqa: E402
from opentrons.types import Point  # noqa: E402
from opentrons_shared_data.deck import load as load_deck_def  # noqa: E402


class InvalidInput(Exception):
    """Invalid input exception."""

    pass


VERSION = 2.0
MOUNT = OT3Mount.GRIPPER
deck_def = load_deck_def("ot3_standard", version=3)
# Offset of the center of the gripper jaw to the desired location
GRIPPER_OFFSET = Point(0.0, 1.0, 0.0)
AVAILABLE_SLOTS = [1, 3, 4, 9, 10, 12]


class GripperState(enum.Enum):
    UNGRIPPING = enum.auto()
    GRIPPING = enum.auto()
    Z_HOMED = enum.auto()
    Z_LOWERED = enum.auto()


VERSION = 2.0
MOUNT = OT3Mount.GRIPPER
deck_def = load_deck_def("ot3_standard", version=3)
# Offset of the center of the gripper jaw to the desired location
GRIPPER_OFFSET = Point(0.0, 1.0, 0.0)
AVAILABLE_SLOTS = [1, 3, 4, 9, 10, 12]


def prompt_int_input(prompt_name: str) -> int:
    """Prompt to choose a member of the enum.

    Args:
        output_func: Function to output text to user.
        get_user_input: Function to get user input.
        enum_type: an enum type

    Returns:
        The choice.

    """
    try:
        return int(input(f"{prompt_name}: "))
    except (ValueError, IndexError) as e:
        raise InvalidInput(e)


def prompt_float_input(prompt_name: str) -> float:
    """Prompt for a float."""
    try:
        return float(input(f"{prompt_name}: "))
    except (ValueError, IndexError) as e:
        raise InvalidInput(e)


def prompt_bool_input(prompt_name: str) -> float:
    """Prompt for a bool."""
    try:
        return {"true": True, "false": False}[input(f"{prompt_name}: ").lower()]
    except (ValueError, IndexError) as e:
        raise InvalidInput(e)


def get_slot_center_in_deck_coord(slot_id: int) -> Point:
    corner = Point(*deck_def["locations"]["orderedSlots"][slot_id - 1]["position"])
    return Point(corner.x + 128.0 / 2, corner.y + 86.0 / 2, corner.z)


def build_api() -> ThreadManager[HardwareControlAPI]:
    api = ThreadManager(OT3API.build_hardware_controller)
    api.managed_thread_ready_blocking()
    return api


def print_current_state(
    api: ThreadManager[HardwareControlAPI],
    cycle_index: int,
    slot: int,
    jaw_state: GripperState,
) -> None:
    """
    1. timestamp
    2. current cycle
    3. current slot
    4. gripper position (X, Y, Z_G)
    5. gripper state: open/closed
    6. encoder positions (X, Y, G)
    """
    pos = api.sync.current_position_ot3(OT3Mount.GRIPPER)
    gripper_loc = (pos[Axis.X], pos[Axis.Y], pos[Axis.Z_G])

    enc_pos = api.sync._encoder_position
    enc_loc = (enc_pos[Axis.X], enc_pos[Axis.Y], enc_pos[Axis.G])
    print(
        f"{datetime.datetime.now()}, {cycle_index}, {slot}, "
        f"{gripper_loc}, {jaw_state}, {enc_loc}\n"
    )


def add_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "--origin",
        type=int,
        required=False,
        default=1,
        help="select slot from 1, 3, 4, 9, 10, 12",
    )
    parser.add_argument(
        "--grip_force",
        type=float,
        required=False,
        default=20,
        help="grip force in Newton",
    )
    parser.add_argument(
        "--grip_height",
        type=float,
        required=False,
        default=25,
        help="grip height in mm",
    )
    parser.add_argument(
        "--repeats",
        type=int,
        required=False,
        default=0,
        help="how many times to repeat the test",
    )
    return parser


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Gripper control script.")
    add_args(parser)
    args = parser.parse_args()

    hc_api = build_api()

    from_slot = args.origin
    grip_force = args.grip_force
    grip_height = args.grip_height
    repeats = args.repeats

    api = hc_api.sync
    api.home()
    homed_pos = api.gantry_position(MOUNT)

    from_slot_loc = get_slot_center_in_deck_coord(from_slot) + GRIPPER_OFFSET

    AVAILABLE_SLOTS.remove(from_slot)

    for cycle in range(repeats + 1):
        for slot_index, to_slot in enumerate(AVAILABLE_SLOTS):
            to_slot_loc = get_slot_center_in_deck_coord(to_slot) + GRIPPER_OFFSET
            if cycle == 0 and slot_index == 0:
                # move to pick up position at origin slot
                api.move_to(MOUNT, from_slot_loc._replace(z=homed_pos.z))
                print_current_state(hc_api, cycle, from_slot, GripperState.Z_HOMED)
                api.move_to(MOUNT, from_slot_loc._replace(z=grip_height))
                print_current_state(hc_api, cycle, from_slot, GripperState.Z_LOWERED)

            api.grip(grip_force)
            api.delay(1)
            print_current_state(hc_api, cycle, from_slot, GripperState.GRIPPING)

            # move to drop off position at destination slot
            api.move_to(MOUNT, from_slot_loc._replace(z=homed_pos.z))
            print_current_state(hc_api, cycle, from_slot, GripperState.Z_HOMED)
            api.move_to(MOUNT, to_slot_loc._replace(z=homed_pos.z))
            print_current_state(hc_api, cycle, to_slot, GripperState.Z_HOMED)
            api.move_to(MOUNT, to_slot_loc._replace(z=grip_height))
            print_current_state(hc_api, cycle, to_slot, GripperState.Z_LOWERED)

            api.ungrip()
            api.delay(1)
            print_current_state(hc_api, cycle, to_slot, GripperState.UNGRIPPING)

            api.grip(grip_force)
            api.delay(1.0)
            print_current_state(hc_api, cycle, to_slot, GripperState.GRIPPING)

            api.move_to(MOUNT, to_slot_loc._replace(z=homed_pos.z))
            print_current_state(hc_api, cycle, to_slot, GripperState.Z_HOMED)
            api.move_to(MOUNT, from_slot_loc._replace(z=homed_pos.z))
            print_current_state(hc_api, cycle, from_slot, GripperState.Z_HOMED)
            api.move_to(MOUNT, from_slot_loc._replace(z=grip_height))
            print_current_state(hc_api, cycle, from_slot, GripperState.Z_LOWERED)

            api.ungrip()
            api.delay(1.0)
            print_current_state(hc_api, cycle, from_slot, GripperState.UNGRIPPING)

    api.home()



"""Command models to start heating a Temperature Module."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

SetTargetTemperatureCommandType = Literal["temperatureModule/setTargetTemperature"]


class SetTargetTemperatureParams(BaseModel):
    """Input parameters to set a Temperature Module's target temperature."""

    moduleId: str = Field(..., description="Unique ID of the Temperature Module.")
    celsius: float = Field(..., description="Target temperature in °C.")


class SetTargetTemperatureResult(BaseModel):
    """Result data from setting a Temperature Module's target temperature."""

    targetTemperature: float = Field(
        ...,
        description="The target temperature that was set after validation "
        "and type conversion (if any).",
    )


class SetTargetTemperatureImpl(
    AbstractCommandImpl[
        SetTargetTemperatureParams, SuccessData[SetTargetTemperatureResult]
    ]
):
    """Execution implementation of a Temperature Module's set temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: SetTargetTemperatureParams
    ) -> SuccessData[SetTargetTemperatureResult]:
        """Set a Temperature Module's target temperature."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        module_substate = self._state_view.modules.get_temperature_module_substate(
            module_id=params.moduleId
        )

        # Verify temperature from temperature module view
        validated_temp = module_substate.validate_target_temperature(params.celsius)

        # Allow propagation of ModuleNotAttachedError.
        temp_hardware_module = self._equipment.get_module_hardware_api(
            module_substate.module_id
        )

        if temp_hardware_module is not None:
            await temp_hardware_module.start_set_temperature(celsius=validated_temp)
        return SuccessData(
            public=SetTargetTemperatureResult(targetTemperature=validated_temp),
        )


class SetTargetTemperature(
    BaseCommand[SetTargetTemperatureParams, SetTargetTemperatureResult, ErrorOccurrence]
):
    """A command to set a Temperature Module's target temperature."""

    commandType: SetTargetTemperatureCommandType = (
        "temperatureModule/setTargetTemperature"
    )
    params: SetTargetTemperatureParams
    result: Optional[SetTargetTemperatureResult]

    _ImplementationCls: Type[SetTargetTemperatureImpl] = SetTargetTemperatureImpl


class SetTargetTemperatureCreate(BaseCommandCreate[SetTargetTemperatureParams]):
    """A request to create a Temperature Module's set temperature command."""

    commandType: SetTargetTemperatureCommandType = (
        "temperatureModule/setTargetTemperature"
    )
    params: SetTargetTemperatureParams

    _CommandCls: Type[SetTargetTemperature] = SetTargetTemperature



from __future__ import annotations

import asyncio
import logging
from typing import Optional, Mapping
from typing_extensions import Final

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.heater_shaker.driver import HeaterShakerDriver
from opentrons.drivers.heater_shaker.abstract import AbstractHeaterShakerDriver
from opentrons.drivers.heater_shaker.simulator import SimulatingDriver
from opentrons.drivers.types import Temperature, RPM, HeaterShakerLabwareLatchStatus
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.poller import Reader, Poller
from opentrons.hardware_control.modules import mod_abc, update
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleType,
    TemperatureStatus,
    SpeedStatus,
    HeaterShakerStatus,
    UploadFunction,
    LiveData,
)

log = logging.getLogger(__name__)

POLL_PERIOD = 1.0

# TODO(mc, 2022-06-14): this techinque copied from temperature module
# to speed up simulation of heater-shaker protocols, but it's pretty silly
# module simulation in PAPIv2 needs to be seriously rethought
SIMULATING_POLL_PERIOD = POLL_PERIOD / 20.0

DFU_PID = "df11"


class HeaterShaker(mod_abc.AbstractModule):
    """Hardware control interface for an attached Heater-Shaker module."""

    MODULE_TYPE = ModuleType.HEATER_SHAKER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "HeaterShaker":
        """
        Build a HeaterShaker

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            HeaterShaker instance
        """
        driver: AbstractHeaterShakerDriver
        if not simulating:
            driver = await HeaterShakerDriver.create(port=port, loop=hw_control_loop)
            poll_interval_seconds = poll_interval_seconds or POLL_PERIOD
        else:
            driver = SimulatingDriver(serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIMULATING_POLL_PERIOD

        reader = HeaterShakerReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=await driver.get_device_info(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Heater-Shaker on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractHeaterShakerDriver,
        reader: HeaterShakerReader,
        poller: Poller,
        device_info: Mapping[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ):
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )
        self._device_info = device_info
        self._driver = driver
        self._reader = reader
        self._poller = poller

    async def cleanup(self) -> None:
        """Stop the poller task"""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        """Used for picking up serial port symlinks"""
        return "heatershaker"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "heater-shaker"

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        return "heaterShakerModuleV1"

    @staticmethod
    def _get_temperature_status(temperature: Temperature) -> TemperatureStatus:
        """
        Determine the status from the temperature.

        Args:
            temperature: A Temperature instance

        Returns:
            The status
        """
        DELTA: Final = 0.7
        status = TemperatureStatus.IDLE
        if temperature.target is not None:
            diff = temperature.target - temperature.current
            if abs(diff) < DELTA:  # To avoid status fluctuation near target
                status = TemperatureStatus.HOLDING
            elif diff < 0:
                status = TemperatureStatus.COOLING
            else:
                status = TemperatureStatus.HEATING
        return status

    @staticmethod
    def _get_speed_status(speed: RPM) -> SpeedStatus:
        """
        Determine the status from the speed.

        Args:
            speed: An RPM instance

        Returns:
            The status
        """
        DELTA: Final = 40
        status = SpeedStatus.IDLE
        if speed.target is not None:
            diff = speed.target - speed.current
            if abs(diff) < DELTA:  # To avoid status fluctuation near target
                status = SpeedStatus.HOLDING
            elif diff < 0:
                status = SpeedStatus.DECELERATING
            else:
                status = SpeedStatus.ACCELERATING
        return status

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    def bootloader(self) -> UploadFunction:
        return update.upload_via_dfu

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def live_data(self) -> LiveData:
        return {
            "status": self.status.value,
            "data": {
                "temperatureStatus": self.temperature_status.value,
                "speedStatus": self.speed_status.value,
                "labwareLatchStatus": self.labware_latch_status.value,
                "currentTemp": self.temperature,
                "targetTemp": self.target_temperature,
                "currentSpeed": self.speed,
                "targetSpeed": self.target_speed,
                "errorDetails": self._reader.error,
            },
        }

    @property
    def temperature(self) -> float:
        return self._reader.temperature.current

    @property
    def target_temperature(self) -> Optional[float]:
        return self._reader.temperature.target

    @property
    def speed(self) -> int:
        return self._reader.rpm.current

    @property
    def target_speed(self) -> Optional[int]:
        return self._reader.rpm.target

    @property
    def temperature_status(self) -> TemperatureStatus:
        return self._get_temperature_status(self._reader.temperature)

    @property
    def speed_status(self) -> SpeedStatus:
        return self._get_speed_status(self._reader.rpm)

    @property
    def labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        return self._reader.labware_latch

    @property
    def status(self) -> HeaterShakerStatus:
        """Module status or error state details."""
        # TODO (spp, 2022-2-22): Does this make sense as the overarching 'status'?
        #  Or maybe consolidate the above 3 statuses into this one?
        if self._reader.error:
            return HeaterShakerStatus.ERROR
        elif (
            self.temperature_status == TemperatureStatus.IDLE
            and self.speed_status == SpeedStatus.IDLE
        ):
            return HeaterShakerStatus.IDLE
        else:
            return HeaterShakerStatus.RUNNING

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    async def start_set_temperature(self, celsius: float) -> None:
        """
        Set temperature in degrees Celsius

        Range: Room temperature to 90 degree Celsius
               Any temperature above the value will be clipped to
               the nearest limit. This is a resistive heater, not
               a Peltier TEC, so a temperature that is too low
               may be unattainable but we do not limit input on the
               low side in case the user has this in a freezer.

        This function will complete as soon as heating begins, and
        will not wait until the temperature is achieved. To then wait
        until heating is complete, use await_temperature. To start
        heating and wait until heating is complete in one call, use
        set_temperature.

        """
        await self.wait_for_is_running()
        await self._driver.set_temperature(celsius)
        await self._reader.read_temperature()

    # TODO(mc, 2022-10-10): remove `awaiting_temperature` argument,
    # and instead, wait until status is holding
    async def await_temperature(self, awaiting_temperature: float) -> None:
        """Await temperature in degrees Celsius.

        Polls the Heater-Shaker's current temperature until
        the specified temperature is reached. If `awaiting_temperature`
        is different than the current target temperature,
        the resulting behavior may be unpredictable.
        """
        if self.is_simulated:
            return

        await self.wait_for_is_running()
        await self._reader.read_temperature()

        async def _await_temperature() -> None:
            if self.temperature_status == TemperatureStatus.HEATING:
                while self.temperature < awaiting_temperature:
                    await self._poller.wait_next_poll()
            elif self.temperature_status == TemperatureStatus.COOLING:
                while self.temperature > awaiting_temperature:
                    await self._poller.wait_next_poll()

        t = self._loop.create_task(_await_temperature())
        self.make_cancellable(t)
        await t

    async def set_speed(self, rpm: int) -> None:
        """
        Set shake speed in RPM

        Range: 0-3000 RPM
               Any speed above or below these values will cause an error.

        This function will not complete until the heater-shaker is at
        the speed or an error occurs. To start spinning but not wait
        until the final speed is reached, see start_set_speed.
        """
        await self.wait_for_is_running()
        await self._driver.set_rpm(rpm)
        await self._reader.read_rpm()

        async def _wait() -> None:
            # Wait until we reach the target speed.
            while self.speed_status != SpeedStatus.HOLDING:
                await self._poller.wait_next_poll()

        task = self._loop.create_task(_wait())
        self.make_cancellable(task)
        await task

    async def _wait_for_labware_latch(
        self, status: HeaterShakerLabwareLatchStatus
    ) -> None:
        """Wait until the hardware reports the labware latch status matches."""
        while self.labware_latch_status != status:
            await self._poller.wait_next_poll()

    async def _wait_for_shake_deactivation(self) -> None:
        """Wait until hardware reports that module has stopped shaking and has homed."""
        while self.speed_status != SpeedStatus.IDLE:
            await self._poller.wait_next_poll()

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Stop heating/cooling; stop shaking and home the plate"""
        await self.deactivate_heater(must_be_running=must_be_running)
        await self.deactivate_shaker(must_be_running=must_be_running)

    async def deactivate_heater(self, must_be_running: bool = True) -> None:
        """Stop heating/cooling"""
        if must_be_running:
            await self.wait_for_is_running()
        await self._driver.deactivate_heater()
        await self._reader.read_temperature()

    async def deactivate_shaker(self, must_be_running: bool = True) -> None:
        """Stop shaking and home the plate"""
        if must_be_running:
            await self.wait_for_is_running()
        await self._driver.home()
        await self._wait_for_shake_deactivation()

    async def open_labware_latch(self) -> None:
        await self.wait_for_is_running()
        await self._driver.open_labware_latch()
        await self._wait_for_labware_latch(HeaterShakerLabwareLatchStatus.IDLE_OPEN)

    async def close_labware_latch(self) -> None:
        await self.wait_for_is_running()
        await self._driver.close_labware_latch()
        await self._wait_for_labware_latch(HeaterShakerLabwareLatchStatus.IDLE_CLOSED)

    async def prep_for_update(self) -> str:
        await self._poller.stop()
        await self._driver.enter_programming_mode()
        dfu_info = await update.find_dfu_device(pid=DFU_PID, expected_device_count=2)
        return dfu_info


class HeaterShakerReader(Reader):
    temperature: Temperature
    rpm: RPM
    labware_latch: HeaterShakerLabwareLatchStatus
    error: Optional[str]

    def __init__(self, driver: AbstractHeaterShakerDriver) -> None:
        self.temperature = Temperature(current=25, target=None)
        self.rpm = RPM(current=0, target=None)
        self.labware_latch = HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN
        self.error: Optional[str] = None
        self._driver = driver

    async def read(self) -> None:
        await self.read_temperature()
        await self.read_rpm()
        await self.read_labware_latch()
        self._set_error(None)

    def on_error(self, exception: Exception) -> None:
        self._set_error(exception)

    async def read_temperature(self) -> None:
        self.temperature = await self._driver.get_temperature()

    async def read_rpm(self) -> None:
        self.rpm = await self._driver.get_rpm()

    async def read_labware_latch(self) -> None:
        self.labware_latch = await self._driver.get_labware_latch_status()

    def _set_error(self, exception: Optional[Exception]) -> None:
        if exception is None:
            self.error = None
        else:
            try:
                self.error = str(exception.args[0])
            except Exception:
                self.error = repr(exception)



import typing

from opentrons.util.helpers import utc_now

from .. import types as local_types

from .models import v1


CalibrationType = typing.TypeVar(
    "CalibrationType",
    v1.PipetteOffsetCalibration,
    v1.TipLengthCalibration,
    v1.DeckCalibrationModel,
)


def mark_bad(
    calibration: CalibrationType,
    source_marked_bad: local_types.SourceType,
) -> CalibrationType:
    calibration.status = v1.CalibrationStatus(
        markedBad=True, source=source_marked_bad, markedAt=utc_now()
    )
    return calibration



from __future__ import annotations

import logging
import json

from pathlib import Path
from typing import Any, AnyStr, Dict, Optional, Union

import jsonschema  # type: ignore

from opentrons_shared_data import load_shared_data, get_shared_data_root
from opentrons.protocols.api_support.constants import (
    OPENTRONS_NAMESPACE,
    CUSTOM_NAMESPACE,
    STANDARD_DEFS_PATH,
    USER_DEFS_PATH,
)
from opentrons_shared_data.labware.types import LabwareDefinition


MODULE_LOG = logging.getLogger(__name__)


def get_labware_definition(
    load_name: str,
    namespace: Optional[str] = None,
    version: Optional[int] = None,
    bundled_defs: Optional[Dict[str, LabwareDefinition]] = None,
    extra_defs: Optional[Dict[str, LabwareDefinition]] = None,
) -> LabwareDefinition:
    """
    Look up and return a definition by load_name + namespace + version and
        return it or raise an exception

    :param str load_name: corresponds to 'loadName' key in definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: A bundle of labware definitions to exclusively use for
        finding labware definitions, if specified
    :param extra_defs: An extra set of definitions (in addition to the system
        definitions) in which to search
    """
    load_name = load_name.lower()

    if bundled_defs is not None:
        return _get_labware_definition_from_bundle(
            bundled_defs, load_name, namespace, version
        )

    checked_extras = extra_defs or {}

    try:
        return _get_labware_definition_from_bundle(
            checked_extras, load_name, namespace, version
        )
    except (FileNotFoundError, RuntimeError):
        pass

    return _get_standard_labware_definition(load_name, namespace, version)


def save_definition(
    labware_def: LabwareDefinition, force: bool = False, location: Optional[Path] = None
) -> None:
    """
    Save a labware definition

    :param labware_def: A deserialized JSON labware definition
    :param bool force: If true, overwrite an existing definition if found.
        Cannot overwrite Opentrons definitions.
    :param location: File path
    """
    namespace = labware_def["namespace"]
    load_name = labware_def["parameters"]["loadName"]
    version = labware_def["version"]

    verify_definition(labware_def)

    if not namespace or not load_name or not version:
        raise RuntimeError(
            "Could not save definition, labware def is missing a field: "
            + f"{namespace}, {load_name}, {version}"
        )

    if namespace == OPENTRONS_NAMESPACE:
        raise RuntimeError(
            f'Saving definitions to the "{OPENTRONS_NAMESPACE}" namespace '
            + "is not permitted"
        )
    def_path = _get_path_to_labware(load_name, namespace, version, location)

    if not force and def_path.is_file():
        raise RuntimeError(
            f"The given definition ({namespace}/{load_name} v{version}) "
            + "already exists. Cannot save definition without force=True"
        )

    Path(def_path).parent.mkdir(parents=True, exist_ok=True)
    with open(def_path, "w") as f:
        json.dump(labware_def, f)


def verify_definition(
    contents: Union[AnyStr, LabwareDefinition, Dict[str, Any]]
) -> LabwareDefinition:
    """Verify that an input string is a labware definition and return it.

    If the definition is invalid, an exception is raised; otherwise parse the
    json and return the valid definition.

    :raises json.JsonDecodeError: If the definition is not valid json
    :raises jsonschema.ValidationError: If the definition is not valid.
    :returns: The parsed definition
    """
    schema_body = load_shared_data("labware/schemas/2.json").decode("utf-8")
    labware_schema_v2 = json.loads(schema_body)

    if isinstance(contents, dict):
        to_return = contents
    else:
        to_return = json.loads(contents)
    jsonschema.validate(to_return, labware_schema_v2)
    # we can type ignore this because if it passes the jsonschema it has
    # the correct structure
    return to_return  # type: ignore[return-value]


def _get_labware_definition_from_bundle(
    bundled_labware: Dict[str, LabwareDefinition],
    load_name: str,
    namespace: Optional[str] = None,
    version: Optional[int] = None,
) -> LabwareDefinition:
    """
    Look up and return a bundled definition by ``load_name`` + ``namespace``
    + ``version`` and return it or raise an exception. The``namespace`` and
    ``version`` args are optional, they only have to be specified if there is
    ambiguity (eg when multiple labware in the bundle share the same
    ``load_name``)

    :param str load_name: corresponds to 'loadName' key in definition
    :param str namespace: The namespace the labware definition belongs to
    :param int version: The version of the labware definition
    :param Dict bundled_labware: A dictionary of labware definitions to search
    """
    load_name = load_name.lower()

    bundled_candidates = [
        b for b in bundled_labware.values() if b["parameters"]["loadName"] == load_name
    ]
    if namespace:
        namespace = namespace.lower()
        bundled_candidates = [
            b for b in bundled_candidates if b["namespace"] == namespace
        ]
    if version:
        bundled_candidates = [b for b in bundled_candidates if b["version"] == version]

    if len(bundled_candidates) == 1:
        return bundled_candidates[0]
    elif len(bundled_candidates) > 1:
        raise RuntimeError(
            f"Ambiguous labware access. Bundle contains multiple "
            f"labware with load name {load_name}, "
            f"namespace {namespace}, and version {version}."
        )
    else:
        raise RuntimeError(
            f"No labware found in bundle with load name {load_name}, "
            f"namespace {namespace}, and version {version}."
        )


def _get_standard_labware_definition(
    load_name: str, namespace: Optional[str] = None, version: Optional[int] = None
) -> LabwareDefinition:

    if version is None:
        checked_version = 1
    else:
        checked_version = version
    error_msg_string = """Unable to find a labware
        definition for "{0}",
        version {1}, in the {2} namespace.
        Please confirm your protocol includes the correct
        labware spelling and (optionally) the correct version
        number and namespace.

        If you are referencing a custom labware in your
        protocol, you must add it to your Custom Labware
        Definitions Folder from the Opentrons App before
        uploading your protocol.
        """
    if namespace is None:
        for fallback_namespace in [OPENTRONS_NAMESPACE, CUSTOM_NAMESPACE]:
            try:
                return _get_standard_labware_definition(
                    load_name, fallback_namespace, checked_version
                )
            except FileNotFoundError:
                pass

        raise FileNotFoundError(
            error_msg_string.format(load_name, checked_version, OPENTRONS_NAMESPACE)
        )

    namespace = namespace.lower()
    def_path = _get_path_to_labware(load_name, namespace, checked_version)

    try:
        with open(def_path, "rb") as f:
            labware_def = json.loads(f.read().decode("utf-8"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f'Labware "{load_name}" not found with version {checked_version} '
            f'in namespace "{namespace}".'
        )
    return labware_def  # type: ignore[no-any-return]


def _get_path_to_labware(
    load_name: str, namespace: str, version: int, base_path: Optional[Path] = None
) -> Path:
    if namespace == OPENTRONS_NAMESPACE:
        # all labware in OPENTRONS_NAMESPACE is stored in shared data
        schema_3_path = (
            get_shared_data_root()
            / STANDARD_DEFS_PATH
            / "3"
            / load_name
            / f"{version}.json"
        )
        schema_2_path = (
            get_shared_data_root()
            / STANDARD_DEFS_PATH
            / "2"
            / load_name
            / f"{version}.json"
        )
        return schema_3_path if schema_3_path.exists() else schema_2_path
    if not base_path:
        base_path = USER_DEFS_PATH
    def_path = base_path / namespace / load_name / f"{version}.json"
    return def_path



import sys
import logging

_pyversion = sys.version_info[0:2]

if _pyversion >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata  # type: ignore[no-redef]

try:
    version: str = metadata.version("opentrons")
except Exception as e:
    logging.warning(
        "Could not determine version for opentrons, may be dev install, using 0.0.0-dev"
    )
    logging.debug(e)
    version = "0.0.0-dev"

__all__ = ["version"]



# Inspired by:
# https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
import sys
import codecs
import os
import os.path

from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, "..", "scripts"))

from python_build_utils import normalize_version  # noqa: E402

# make stdout blocking since Travis sets it to nonblocking
if os.name == "posix":
    import fcntl

    flags = fcntl.fcntl(sys.stdout, fcntl.F_GETFL)
    fcntl.fcntl(sys.stdout, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)


def get_version():
    buildno = os.getenv("BUILD_NUMBER")
    project = os.getenv("OPENTRONS_PROJECT", "robot-stack")
    git_dir = os.getenv("OPENTRONS_GIT_DIR", None)
    if buildno:
        normalize_opts = {"extra_tag": buildno}
    else:
        normalize_opts = {}
    return normalize_version("api", project, git_dir=git_dir, **normalize_opts)


VERSION = get_version()

DISTNAME = "opentrons"
LICENSE = "Apache 2.0"
AUTHOR = "Opentrons"
EMAIL = "engineering@opentrons.com"
URL = "https://github.com/OpenTrons/opentrons"
DOWNLOAD_URL = ""
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.10",
    "Topic :: Scientific/Engineering",
]
KEYWORDS = ["robots", "protocols", "synbio", "pcr", "automation", "lab"]
DESCRIPTION = (
    "The Opentrons API is a simple framework designed to make "
    "writing automated biology lab protocols easy."
)
PACKAGES = find_packages(where="src")
INSTALL_REQUIRES = [
    f"opentrons-shared-data=={VERSION}",
    "aionotify==0.3.1",
    "anyio>=3.6.1,<4.0.0",
    "jsonschema>=3.0.1,<4.18.0",
    "numpy>=1.20.0,<2",
    "pydantic>=1.10.9,<2.0.0",
    "pyserial>=3.5",
    "typing-extensions>=4.0.0,<5",
    "click>=8.0.0,<9",
    "pyusb==1.2.1",
    'importlib-metadata >= 1.0 ; python_version < "3.8"',
    "packaging>=21.0",
]

EXTRAS = {
    "ot2-hardware": [f"opentrons-hardware=={VERSION}"],
    "flex-hardware": [f"opentrons-hardware[FLEX]=={VERSION}"],
}


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


if __name__ == "__main__":
    setup(
        python_requires=">=3.10",
        name=DISTNAME,
        description=DESCRIPTION,
        license=LICENSE,
        version=VERSION,
        author=AUTHOR,
        author_email=EMAIL,
        maintainer=AUTHOR,
        maintainer_email=EMAIL,
        keywords=KEYWORDS,
        long_description=read("pypi-readme.rst"),
        packages=PACKAGES,
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS,
        include_package_data=True,
        package_dir={"": "src"},
        package_data={"opentrons": ["py.typed"]},
        entry_points={
            "console_scripts": [
                "opentrons_simulate = opentrons.simulate:main",
                "opentrons_execute = opentrons.execute:main",
            ]
        },
        project_urls={
            "opentrons.com": "https://www.opentrons.com",
            "Source Code On Github": "https://github.com/Opentrons/opentrons/tree/edge/api",
            "Documentation": "https://docs.opentrons.com",
        },
    )



# noqa: D100


from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.actions.action_handler import ActionHandler
from opentrons.protocol_engine.actions.actions import (
    Action,
    FailCommandAction,
    ResumeFromRecoveryAction,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.execution.tip_handler import HardwareTipHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView


class ErrorRecoveryHardwareStateSynchronizer(ActionHandler):
    """A hack to keep the hardware API's state correct through certain error recovery flows.

    BACKGROUND:

    Certain parts of robot state are duplicated between `opentrons.protocol_engine` and
    `opentrons.hardware_control`. Stuff like "is there a tip attached."

    Normally, Protocol Engine command implementations (`opentrons.protocol_engine.commands`)
    mutate hardware API state when they execute; and then when they finish executing,
    the Protocol Engine state stores (`opentrons.protocol_engine.state`) update Protocol
    Engine state accordingly. So both halves are accounted for. This generally works fine.

    However, we need to go out of our way to support
    `ProtocolEngine.resume_from_recovery(reconcile_false_positive=True)`.
    It wants to apply a second set of state updates to "fix things up" with the
    new knowledge that some error was a false positive. The Protocol Engine half of that
    is easy for us to apply the normal way, through the state stores; but the
    hardware API half of that cannot be applied the normal way, from the command
    implementation, because the command in question is no longer running.

    THE HACK:

    This listens for the same error recovery state updates that the state stores do,
    figures out what hardware API state mutations ought to go along with them,
    and then does those mutations.

    The problem is that hardware API state is now mutated from two different places
    (sometimes the command implementations, and sometimes here), which are bound
    to grow accidental differences.

    TO FIX:

    Make Protocol Engine's use of the hardware API less stateful. e.g. supply
    tip geometry every time we call a hardware API movement method, instead of
    just once when we pick up a tip. Use Protocol Engine state as the single source
    of truth.
    """

    def __init__(self, hardware_api: HardwareControlAPI, state_view: StateView) -> None:
        self._hardware_api = hardware_api
        self._state_view = state_view

    def handle_action(self, action: Action) -> None:
        """Modify hardware API state in reaction to a Protocol Engine action."""
        state_update = _get_state_update(action)
        if state_update:
            self._synchronize(state_update)

    def _synchronize(self, state_update: update_types.StateUpdate) -> None:
        tip_handler = HardwareTipHandler(self._state_view, self._hardware_api)

        if state_update.pipette_tip_state != update_types.NO_CHANGE:
            pipette_id = state_update.pipette_tip_state.pipette_id
            tip_geometry = state_update.pipette_tip_state.tip_geometry
            if tip_geometry is None:
                tip_handler.remove_tip(pipette_id)
            else:
                tip_handler.cache_tip(pipette_id=pipette_id, tip=tip_geometry)


def _get_state_update(action: Action) -> update_types.StateUpdate | None:
    """Get the mutations that we need to do on the hardware API to stay in sync with an engine action.

    The mutations are returned in Protocol Engine terms, as a StateUpdate.
    They then need to be converted to hardware API terms.
    """
    match action:
        case ResumeFromRecoveryAction(state_update=state_update):
            return state_update

        case FailCommandAction(
            error=DefinedErrorData(
                state_update_if_false_positive=state_update_if_false_positive
            )
        ):
            return (
                state_update_if_false_positive
                if action.type == ErrorRecoveryType.ASSUME_FALSE_POSITIVE_AND_CONTINUE
                else None
            )

        case _:
            return None



"""Move to well command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    WellLocationMixin,
    MovementMixin,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..state import update_types
from ..errors import LabwareIsTipRackError

if TYPE_CHECKING:
    from ..execution import MovementHandler
    from ..state.state import StateView

MoveToWellCommandType = Literal["moveToWell"]


class MoveToWellParams(PipetteIdMixin, WellLocationMixin, MovementMixin):
    """Payload required to move a pipette to a specific well."""

    pass


class MoveToWellResult(DestinationPositionResult):
    """Result data from the execution of a MoveToWell command."""

    pass


class MoveToWellImplementation(
    AbstractCommandImpl[MoveToWellParams, SuccessData[MoveToWellResult]]
):
    """Move to well command implementation."""

    def __init__(
        self, state_view: StateView, movement: MovementHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._movement = movement

    async def execute(self, params: MoveToWellParams) -> SuccessData[MoveToWellResult]:
        """Move the requested pipette to the requested well."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        well_location = params.wellLocation

        state_update = update_types.StateUpdate()

        if (
            self._state_view.labware.is_tiprack(labware_id)
            and well_location.volumeOffset
        ):
            raise LabwareIsTipRackError(
                "Cannot specify a WellLocation with a volumeOffset with movement to a tip rack"
            )

        x, y, z = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            force_direct=params.forceDirect,
            minimum_z_height=params.minimumZHeight,
            speed=params.speed,
        )
        deck_point = DeckPoint.construct(x=x, y=y, z=z)
        state_update.set_pipette_location(
            pipette_id=pipette_id,
            new_labware_id=labware_id,
            new_well_name=well_name,
            new_deck_point=deck_point,
        )

        return SuccessData(
            public=MoveToWellResult(position=deck_point),
            state_update=state_update,
        )


class MoveToWell(BaseCommand[MoveToWellParams, MoveToWellResult, ErrorOccurrence]):
    """Move to well command model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams
    result: Optional[MoveToWellResult]

    _ImplementationCls: Type[MoveToWellImplementation] = MoveToWellImplementation


class MoveToWellCreate(BaseCommandCreate[MoveToWellParams]):
    """Move to well command creation request model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams

    _CommandCls: Type[MoveToWell] = MoveToWell



"""Definitions of data and interface shapes for notes."""
from typing import Union, Literal, Protocol, List, TYPE_CHECKING
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType

NoteKind = Union[Literal["warning", "information"], str]


class CommandNote(BaseModel):
    """A note about a command's execution or dispatch."""

    noteKind: NoteKind = Field(
        ...,
        description="The kind of note this is. Only the literal possibilities should be"
        " relied upon programmatically.",
    )
    shortMessage: str = Field(
        ...,
        description="The accompanying human-readable short message (suitable for display in a single line)",
    )
    longMessage: str = Field(
        ...,
        description="A longer message that may contain newlines and formatting characters describing the note.",
    )
    source: str = Field(
        ..., description="An identifier for the party that created the note"
    )


def make_error_recovery_debug_note(type: "ErrorRecoveryType") -> CommandNote:
    """Return a note for debugging error recovery.

    This is intended to be read by developers and support people, not computers.
    """
    message = f"Handling this command failure with {type.name}."
    return CommandNote.construct(
        noteKind="debugErrorRecovery",
        shortMessage=message,
        longMessage=message,
        source="execution",
    )


class CommandNoteAdder(Protocol):
    """The shape of a function that something can use to add a command note."""

    def __call__(self, note: CommandNote) -> None:
        """When called, this function should add the passed Note to some list."""
        ...


class CommandNoteTracker(CommandNoteAdder, Protocol):
    """The shape of a class that can track notes."""

    def get_notes(self) -> List[CommandNote]:
        """When called, should return all notes previously added with __call__."""
        ...



"""Move relative (jog) command payload, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal


from ..state import update_types
from ..types import MovementAxis, DeckPoint
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from .pipetting_common import DestinationPositionResult

if TYPE_CHECKING:
    from ..execution import MovementHandler


MoveRelativeCommandType = Literal["moveRelative"]


class MoveRelativeParams(BaseModel):
    """Payload required for a MoveRelative command."""

    pipetteId: str = Field(..., description="Pipette to move.")
    axis: MovementAxis = Field(..., description="Axis along which to move.")
    distance: float = Field(
        ...,
        description=(
            "Distance to move in millimeters. A positive number will move"
            " towards the right (x), back (y), top (z) of the deck."
        ),
    )


class MoveRelativeResult(DestinationPositionResult):
    """Result data from the execution of a MoveRelative command."""

    pass


class MoveRelativeImplementation(
    AbstractCommandImpl[MoveRelativeParams, SuccessData[MoveRelativeResult]]
):
    """Move relative command implementation."""

    def __init__(self, movement: MovementHandler, **kwargs: object) -> None:
        self._movement = movement

    async def execute(
        self, params: MoveRelativeParams
    ) -> SuccessData[MoveRelativeResult]:
        """Move (jog) a given pipette a relative distance."""
        state_update = update_types.StateUpdate()

        x, y, z = await self._movement.move_relative(
            pipette_id=params.pipetteId,
            axis=params.axis,
            distance=params.distance,
        )
        deck_point = DeckPoint.construct(x=x, y=y, z=z)
        state_update.pipette_location = update_types.PipetteLocationUpdate(
            pipette_id=params.pipetteId,
            # TODO(jbl 2023-02-14): Need to investigate whether move relative should clear current location
            new_location=update_types.NO_CHANGE,
            new_deck_point=deck_point,
        )

        return SuccessData(
            public=MoveRelativeResult(position=deck_point),
            state_update=state_update,
        )


class MoveRelative(
    BaseCommand[MoveRelativeParams, MoveRelativeResult, ErrorOccurrence]
):
    """Command to move (jog) a given pipette a relative distance."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams
    result: Optional[MoveRelativeResult]

    _ImplementationCls: Type[MoveRelativeImplementation] = MoveRelativeImplementation


class MoveRelativeCreate(BaseCommandCreate[MoveRelativeParams]):
    """Data to create a MoveRelative command."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams

    _CommandCls: Type[MoveRelative] = MoveRelative



"""An emulation of the opentrons heater shaker module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""
import logging
from time import sleep
from typing import (
    Optional,
)

from opentrons.drivers.heater_shaker.driver import (
    GCODE,
    HS_ACK,
)
from opentrons.hardware_control.emulation.parser import Parser, Command
from opentrons.hardware_control.emulation.settings import HeaterShakerSettings
from . import util

from .abstract_emulator import AbstractEmulator
from .simulations import (
    Temperature,
    RPM,
)
from .util import TEMPERATURE_ROOM
from ...drivers.types import HeaterShakerLabwareLatchStatus

logger = logging.getLogger(__name__)


class HeaterShakerEmulator(AbstractEmulator):
    """Heater Shaker emulator"""

    _temperature: Temperature
    _rpm: RPM
    _latch_status: HeaterShakerLabwareLatchStatus

    def __init__(self, parser: Parser, settings: HeaterShakerSettings) -> None:
        self._parser = parser
        self._settings = settings
        self._gcode_to_function_mapping = {
            GCODE.SET_RPM.value: self._set_rpm,
            GCODE.GET_RPM.value: self._get_rpm,
            GCODE.SET_TEMPERATURE.value: self._set_temp,
            GCODE.GET_TEMPERATURE.value: self._get_temp,
            GCODE.HOME.value: self._home,
            GCODE.ENTER_BOOTLOADER.value: self._enter_bootloader,
            GCODE.GET_VERSION.value: self._get_version,
            GCODE.OPEN_LABWARE_LATCH.value: self._open_labware_latch,
            GCODE.CLOSE_LABWARE_LATCH.value: self._close_labware_latch,
            GCODE.GET_LABWARE_LATCH_STATE.value: self._get_labware_latch_state,
            GCODE.DEACTIVATE_HEATER.value: self._deactivate_heater,
        }
        self.reset()

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = " ".join(r for r in results if r)
        return None if not joined else joined

    def reset(self) -> None:

        self._temperature = Temperature(
            per_tick=self._settings.temperature.degrees_per_tick,
            current=self._settings.temperature.starting,
        )
        self._rpm = RPM(
            per_tick=self._settings.rpm.rpm_per_tick,
            current=self._settings.rpm.starting,
        )
        self._rpm.set_target(0.0)
        self._latch_status = HeaterShakerLabwareLatchStatus.IDLE_OPEN

    def _handle(self, command: Command) -> Optional[str]:
        """
        Handle a command.

        TODO: AL 20210218 create dispatch map and remove 'noqa(C901)'
        """
        logger.info(f"Got command {command}")
        func_to_run = self._gcode_to_function_mapping.get(command.gcode)
        res = None if func_to_run is None else func_to_run(command)
        return None if not isinstance(res, str) else f"{res} {HS_ACK}"

    def _set_rpm(self, command: Command) -> str:
        value = command.params["S"]
        assert isinstance(value, float), f"invalid value '{value}'"
        self._rpm.set_target(value)
        return "M3"

    def _get_rpm(self, command: Command) -> str:
        res = (
            f"M123 C:{self._rpm.current} "
            f"T:{self._rpm.target if self._rpm.target is not None else 0}"
        )
        self._rpm.tick()
        return res

    def _set_temp(self, command: Command) -> str:
        value = command.params["S"]
        assert isinstance(value, float), f"invalid value '{value}'"
        self._temperature.set_target(value)
        return "M104"

    def _get_temp(self, command: Command) -> str:
        res = (
            f"M105 C:{self._temperature.current} "
            f"T:{util.OptionalValue(self._temperature.target)}"
        )
        self._temperature.tick()
        return res

    def _home(self, command: Command) -> str:
        sleep(self._settings.home_delay_time)
        self._rpm.deactivate(0.0)
        self._rpm.set_target(0.0)
        return "G28"

    def _enter_bootloader(self, command: Command) -> None:
        pass

    def _get_version(self, command: Command) -> str:
        return (
            f"FW:{self._settings.version} "
            f"HW:{self._settings.model} "
            f"SerialNo:{self._settings.serial_number}"
        )

    def _open_labware_latch(self, command: Command) -> str:
        self._latch_status = HeaterShakerLabwareLatchStatus.IDLE_OPEN
        return "M242"

    def _close_labware_latch(self, command: Command) -> str:
        self._latch_status = HeaterShakerLabwareLatchStatus.IDLE_CLOSED
        return "M243"

    def _get_labware_latch_state(self, command: Command) -> str:
        return f"M241 STATUS:{self._latch_status.value.upper()}"

    def _deactivate_heater(self, command: Command) -> str:
        self._temperature.deactivate(TEMPERATURE_ROOM)
        return "M106"

    @staticmethod
    def get_terminator() -> bytes:
        return b"\n"



"""Command models for Thermocycler commands."""

from .set_target_block_temperature import (
    SetTargetBlockTemperatureCommandType,
    SetTargetBlockTemperatureParams,
    SetTargetBlockTemperatureResult,
    SetTargetBlockTemperature,
    SetTargetBlockTemperatureCreate,
)

from .wait_for_block_temperature import (
    WaitForBlockTemperatureCommandType,
    WaitForBlockTemperatureParams,
    WaitForBlockTemperatureResult,
    WaitForBlockTemperature,
    WaitForBlockTemperatureCreate,
)

from .set_target_lid_temperature import (
    SetTargetLidTemperatureCommandType,
    SetTargetLidTemperatureParams,
    SetTargetLidTemperatureResult,
    SetTargetLidTemperature,
    SetTargetLidTemperatureCreate,
)

from .wait_for_lid_temperature import (
    WaitForLidTemperatureCommandType,
    WaitForLidTemperatureParams,
    WaitForLidTemperatureResult,
    WaitForLidTemperature,
    WaitForLidTemperatureCreate,
)

from .deactivate_block import (
    DeactivateBlockCommandType,
    DeactivateBlockParams,
    DeactivateBlockResult,
    DeactivateBlock,
    DeactivateBlockCreate,
)

from .deactivate_lid import (
    DeactivateLidCommandType,
    DeactivateLidParams,
    DeactivateLidResult,
    DeactivateLid,
    DeactivateLidCreate,
)

from .open_lid import (
    OpenLidCommandType,
    OpenLidParams,
    OpenLidResult,
    OpenLid,
    OpenLidCreate,
)

from .close_lid import (
    CloseLidCommandType,
    CloseLidParams,
    CloseLidResult,
    CloseLid,
    CloseLidCreate,
)

from .run_profile import (
    RunProfileCommandType,
    RunProfileParams,
    RunProfileStepParams,
    RunProfileResult,
    RunProfile,
    RunProfileCreate,
)

from .run_extended_profile import (
    RunExtendedProfileCommandType,
    RunExtendedProfileParams,
    RunExtendedProfileResult,
    RunExtendedProfile,
    RunExtendedProfileCreate,
    ProfileCycle,
    ProfileStep,
)


__all__ = [
    # Set target block temperature command models
    "SetTargetBlockTemperatureCommandType",
    "SetTargetBlockTemperatureParams",
    "SetTargetBlockTemperatureResult",
    "SetTargetBlockTemperature",
    "SetTargetBlockTemperatureCreate",
    # Wait for block temperature command models
    "WaitForBlockTemperatureCommandType",
    "WaitForBlockTemperatureParams",
    "WaitForBlockTemperatureResult",
    "WaitForBlockTemperature",
    "WaitForBlockTemperatureCreate",
    # Set target lid temperature command models
    "SetTargetLidTemperatureCommandType",
    "SetTargetLidTemperatureParams",
    "SetTargetLidTemperatureResult",
    "SetTargetLidTemperature",
    "SetTargetLidTemperatureCreate",
    # Wait for lid temperature command models
    "WaitForLidTemperatureCommandType",
    "WaitForLidTemperatureParams",
    "WaitForLidTemperatureResult",
    "WaitForLidTemperature",
    "WaitForLidTemperatureCreate",
    # Deactivate block command models
    "DeactivateBlockCommandType",
    "DeactivateBlockParams",
    "DeactivateBlockResult",
    "DeactivateBlock",
    "DeactivateBlockCreate",
    # Deactivate lid command models
    "DeactivateLidCommandType",
    "DeactivateLidParams",
    "DeactivateLidResult",
    "DeactivateLid",
    "DeactivateLidCreate",
    # Open lid command models
    "OpenLidCommandType",
    "OpenLidParams",
    "OpenLidResult",
    "OpenLid",
    "OpenLidCreate",
    # Close lid command models
    "CloseLidCommandType",
    "CloseLidParams",
    "CloseLidResult",
    "CloseLid",
    "CloseLidCreate",
    # Run profile command models,
    "RunProfileCommandType",
    "RunProfileParams",
    "RunProfileStepParams",
    "RunProfileResult",
    "RunProfile",
    "RunProfileCreate",
    # Run extended profile command models.
    "RunExtendedProfileCommandType",
    "RunExtendedProfileParams",
    "RunExtendedProfileStepParams",
    "RunExtendedProfileResult",
    "RunExtendedProfile",
    "RunExtendedProfileCreate",
    "ProfileCycle",
    "ProfileStep",
]



from typing import List, Union

from opentrons.protocol_api.labware import Well, Labware
from opentrons.protocol_api.module_contexts import ModuleContext
from opentrons.protocol_api.disposal_locations import TrashBin, WasteChute
from opentrons.protocol_api._types import OffDeckType
from opentrons.types import Location, DeckLocation


CommandLocation = Union[Location, Well]


def listify(
    location: Union[CommandLocation, List[CommandLocation]]
) -> List[CommandLocation]:
    if isinstance(location, list):
        try:
            return listify(location[0])
        except IndexError:
            # TODO(mc, 2021-10-20): this looks like a bug; should this
            # return an empty list, instead?
            return [location]  # type: ignore[list-item]
    else:
        return [location]


def _stringify_new_loc(loc: CommandLocation) -> str:
    if isinstance(loc, Location):
        if loc.labware.is_empty:
            return str(loc.point)
        else:
            return repr(loc.labware)
    elif isinstance(loc, Well):
        return str(loc)
    else:
        raise TypeError(loc)


def stringify_location(location: Union[CommandLocation, List[CommandLocation]]) -> str:
    loc_str_list = [_stringify_new_loc(loc) for loc in listify(location)]
    return ", ".join(loc_str_list)


def stringify_disposal_location(location: Union[TrashBin, WasteChute]) -> str:
    if isinstance(location, TrashBin):
        return f"Trash Bin on slot {location.location.id}"
    elif isinstance(location, WasteChute):
        return "Waste Chute"


def _stringify_labware_movement_location(
    location: Union[
        DeckLocation, OffDeckType, Labware, ModuleContext, WasteChute, TrashBin
    ]
) -> str:
    if isinstance(location, (int, str)):
        return f"slot {location}"
    elif isinstance(location, OffDeckType):
        return "off-deck"
    elif isinstance(location, Labware):
        return location.name
    elif isinstance(location, ModuleContext):
        return str(location)
    elif isinstance(location, WasteChute):
        return "Waste Chute"
    elif isinstance(location, TrashBin):
        return "Trash Bin " + location.location.name


def stringify_labware_movement_command(
    source_labware: Labware,
    destination: Union[
        DeckLocation, OffDeckType, Labware, ModuleContext, WasteChute, TrashBin
    ],
    use_gripper: bool,
) -> str:
    source_labware_text = _stringify_labware_movement_location(source_labware)
    destination_text = _stringify_labware_movement_location(destination)
    gripper_text = " with gripper" if use_gripper else ""
    return f"Moving {source_labware_text} to {destination_text}{gripper_text}"



"""Parameter definition and associated validators."""
from abc import abstractmethod, ABC
from typing import Generic, Optional, List, Set, Union

from opentrons.protocols.parameters.types import (
    ParamType,
    ParameterChoice,
    PrimitiveAllowedTypes,
)
from opentrons.protocols.parameters.exceptions import (
    ParameterValueError,
    ParameterDefinitionError,
)
from opentrons.protocols.parameters import validation
from opentrons.protocol_engine.types import (
    RunTimeParameter,
    NumberParameter,
    BooleanParameter,
    EnumParameter,
    EnumChoice,
)
from opentrons.util.get_union_elements import get_union_elements


class AbstractParameterDefinition(ABC, Generic[ParamType]):
    @property
    @abstractmethod
    def variable_name(self) -> str:
        ...

    @property
    @abstractmethod
    def value(self) -> ParamType:
        ...

    @value.setter
    @abstractmethod
    def value(self, new_value: ParamType) -> None:
        ...

    @abstractmethod
    def as_protocol_engine_type(self) -> RunTimeParameter:
        ...


class ParameterDefinition(AbstractParameterDefinition[PrimitiveAllowedTypes]):
    """The definition for a user defined parameter."""

    def __init__(
        self,
        display_name: str,
        variable_name: str,
        parameter_type: type,
        default: PrimitiveAllowedTypes,
        minimum: Optional[PrimitiveAllowedTypes] = None,
        maximum: Optional[PrimitiveAllowedTypes] = None,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Initializes a parameter.

        This stores the type, default values, range or list of possible values, and other information
        that is defined when a parameter is created for a protocol, as well as validators for setting
        a non-default value for the parameter.

        Arguments:
            display_name: The display name of the parameter as it would show up on the frontend.
            variable_name: The variable name the parameter will be referred to in the run context.
            parameter_type: Can be bool, int, float or str. Must match the type of default and all choices or
                min and max values
            default: The default value the parameter is set to. This will be used in initial analysis.
            minimum: The minimum value the parameter can be set to (inclusive). Mutually exclusive with choices.
            maximum: The maximum value the parameter can be set to (inclusive). Mutually exclusive with choices.
            choices: A sequence of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: An optional description for the parameter.
            unit: An optional suffix for float and int type parameters.
        """
        self._display_name = validation.ensure_display_name(display_name)
        self._variable_name = validation.ensure_variable_name(variable_name)
        self._description = validation.ensure_description(description)
        self._unit = validation.ensure_unit_string_length(unit)

        if parameter_type not in get_union_elements(PrimitiveAllowedTypes):
            raise ParameterDefinitionError(
                "Parameters can only be of type int, float, str, or bool."
            )
        self._type = parameter_type

        self._choices: Optional[List[ParameterChoice]] = choices
        self._allowed_values: Optional[Set[PrimitiveAllowedTypes]] = None

        self._minimum: Optional[Union[int, float]] = None
        self._maximum: Optional[Union[int, float]] = None

        validation.validate_options(default, minimum, maximum, choices, parameter_type)
        if choices is not None:
            self._allowed_values = {choice["value"] for choice in choices}
        else:
            assert isinstance(minimum, (int, float)) and isinstance(
                maximum, (int, float)
            )
            self._minimum = minimum
            self._maximum = maximum

        self._default: PrimitiveAllowedTypes = default
        self.value: PrimitiveAllowedTypes = default

    @property
    def value(self) -> PrimitiveAllowedTypes:
        """The current value of the parameter."""
        return self._value

    @value.setter
    def value(self, new_value: PrimitiveAllowedTypes) -> None:
        validation.validate_type(new_value, self._type)
        if self._allowed_values is not None and new_value not in self._allowed_values:
            raise ParameterValueError(
                f"Parameter must be set to one of the allowed values of {sorted(self._allowed_values)}."
            )
        elif (
            isinstance(self._minimum, (int, float))
            and isinstance(self._maximum, (int, float))
            and isinstance(new_value, (int, float))
            and not (self._minimum <= new_value <= self._maximum)
        ):
            raise ParameterValueError(
                f"Parameter must be between {self._minimum} and {self._maximum} inclusive."
            )
        self._value = new_value

    @property
    def variable_name(self) -> str:
        """The in-protocol variable name of the parameter."""
        return self._variable_name

    @property
    def parameter_type(self) -> type:
        """The python type of the parameter."""
        return self._type

    def as_protocol_engine_type(self) -> RunTimeParameter:
        """Returns parameter as a Protocol Engine type to send to client."""
        parameter: RunTimeParameter
        if self._type is bool:
            parameter = BooleanParameter(
                displayName=self._display_name,
                variableName=self._variable_name,
                description=self._description,
                value=bool(self._value),
                default=bool(self._default),
            )
        elif self._choices is not None:
            choices = [
                EnumChoice(
                    displayName=str(choice["display_name"]),
                    value=choice["value"],
                )
                for choice in self._choices
            ]
            parameter = EnumParameter(
                type=validation.convert_type_string_for_enum(self._type),
                displayName=self._display_name,
                variableName=self._variable_name,
                description=self._description,
                choices=choices,
                value=self._value,
                default=self._default,
            )
        elif self._minimum is not None and self._maximum is not None:
            parameter = NumberParameter(
                type=validation.convert_type_string_for_num_param(self._type),
                displayName=self._display_name,
                variableName=self._variable_name,
                description=self._description,
                suffix=self._unit,
                min=float(self._minimum),
                max=float(self._maximum),
                value=float(self._value),
                default=float(self._default),
            )
        else:
            raise ParameterDefinitionError(
                f"Cannot resolve parameter {self._display_name} to protocol engine type."
            )

        return parameter


def create_int_parameter(
    display_name: str,
    variable_name: str,
    default: int,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    choices: Optional[List[ParameterChoice]] = None,
    description: Optional[str] = None,
    unit: Optional[str] = None,
) -> ParameterDefinition:
    """Creates an integer parameter."""
    return ParameterDefinition(
        parameter_type=int,
        display_name=display_name,
        variable_name=variable_name,
        default=default,
        minimum=minimum,
        maximum=maximum,
        choices=choices,
        description=description,
        unit=unit,
    )


def create_float_parameter(
    display_name: str,
    variable_name: str,
    default: float,
    minimum: Optional[float] = None,
    maximum: Optional[float] = None,
    choices: Optional[List[ParameterChoice]] = None,
    description: Optional[str] = None,
    unit: Optional[str] = None,
) -> ParameterDefinition:
    """Creates a float parameter."""
    return ParameterDefinition(
        parameter_type=float,
        display_name=display_name,
        variable_name=variable_name,
        default=default,
        minimum=minimum,
        maximum=maximum,
        choices=choices,
        description=description,
        unit=unit,
    )


def create_bool_parameter(
    display_name: str,
    variable_name: str,
    default: bool,
    choices: List[ParameterChoice],
    description: Optional[str] = None,
) -> ParameterDefinition:
    """Creates a boolean parameter."""
    return ParameterDefinition(
        parameter_type=bool,
        display_name=display_name,
        variable_name=variable_name,
        default=default,
        choices=choices,
        description=description,
    )


def create_str_parameter(
    display_name: str,
    variable_name: str,
    default: str,
    choices: Optional[List[ParameterChoice]] = None,
    description: Optional[str] = None,
) -> ParameterDefinition:
    """Creates a string parameter."""
    return ParameterDefinition(
        parameter_type=str,
        display_name=display_name,
        variable_name=variable_name,
        default=default,
        choices=choices,
        description=description,
    )



from typing_extensions import Protocol
from ..types import PauseType


class ExecutionControllable(Protocol):
    """A protocol specifying run control (pause, resume)."""

    def pause(self, pause_type: PauseType) -> None:
        """
        Pause motion of the robot after a current motion concludes.

        Individual calls to move
        (which aspirate and dispense and other calls may depend on) are
        considered atomic and will always complete if they have been called
        prior to a call to this method. However, subsequent calls to move that
        occur when the system is paused will not proceed until the system is
        resumed with resume.
        """
        ...

    def pause_with_message(self, message: str) -> None:
        """Pause motion of the robot as with pause, but specify a message."""
        ...

    def resume(self, pause_type: PauseType) -> None:
        """
        Resume motion after a call to pause.
        """
        ...

    async def delay(self, duration_s: float) -> None:
        """Delay execution by pausing and sleeping."""
        ...



""" opentrons.util.entrypoint_util: functions common to entrypoints
"""

import asyncio
import contextlib
from dataclasses import dataclass
import json
import logging
from json import JSONDecodeError
import pathlib
import subprocess
import sys
import tempfile
from typing import (
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Union,
    TYPE_CHECKING,
)

from jsonschema import ValidationError  # type: ignore

from opentrons.calibration_storage.deck_configuration import (
    deserialize_deck_configuration,
)
from opentrons.config import (
    ARCHITECTURE,
    IS_ROBOT,
    JUPYTER_NOTEBOOK_LABWARE_DIR,
    SystemArchitecture,
)
from opentrons.protocol_api import labware
from opentrons.calibration_storage import helpers
from opentrons.protocol_engine.errors.error_occurrence import (
    ErrorOccurrence as ProtocolEngineErrorOccurrence,
)
from opentrons.protocol_engine.types import DeckConfigurationType
from opentrons.protocol_reader import ProtocolReader, ProtocolSource
from opentrons.protocols.types import JsonProtocol, Protocol, PythonProtocol

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition


log = logging.getLogger(__name__)


@dataclass
class FoundLabware:
    """An individual labware found by `labware_from_paths()`."""

    path: pathlib.Path
    definition: "LabwareDefinition"


def labware_from_paths(
    paths: Sequence[Union[str, pathlib.Path]]
) -> Dict[str, FoundLabware]:
    """Search paths for labware definitions.

    Returns:
        A dict, keyed by labware URI, where each value has the file path and the parsed def.
    """
    labware_defs: Dict[str, FoundLabware] = {}

    for strpath in paths:
        log.info(f"local labware: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if not path.is_dir():
            raise RuntimeError(f"{path} is not a directory")
        for child in path.iterdir():
            if child.is_file() and child.suffix.endswith("json"):
                try:
                    defn = labware.verify_definition(child.read_bytes())
                except (ValidationError, JSONDecodeError):
                    log.info(f"{child}: invalid labware, ignoring")
                    log.debug(
                        f"{child}: labware invalid because of this exception.",
                        exc_info=True,
                    )
                else:
                    uri = helpers.uri_from_definition(defn)
                    labware_defs[uri] = FoundLabware(path=child, definition=defn)
                    log.info(f"loaded labware {uri} from {child}")
            else:
                log.info(f"ignoring {child} in labware path")
    return labware_defs


def find_jupyter_labware() -> Optional[Dict[str, FoundLabware]]:
    """Return labware files in this robot's Jupyter Notebook directory.

    Returns:
        If we're running on an Opentrons robot:
        A dict, keyed by labware URI, where each value has the file path and the parsed def.

        Otherwise: None.
    """
    if IS_ROBOT:
        # JUPYTER_NOTEBOOK_LABWARE_DIR should never be None when IS_ROBOT == True.
        assert JUPYTER_NOTEBOOK_LABWARE_DIR is not None
        if JUPYTER_NOTEBOOK_LABWARE_DIR.is_dir():
            return labware_from_paths([JUPYTER_NOTEBOOK_LABWARE_DIR])

    return None


def datafiles_from_paths(paths: Sequence[Union[str, pathlib.Path]]) -> Dict[str, bytes]:
    datafiles: Dict[str, bytes] = {}
    for strpath in paths:
        log.info(f"data files: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if path.is_file():
            datafiles[path.name] = path.read_bytes()
            log.info(f"read {path} into custom data as {path.name}")
        elif path.is_dir():
            for child in path.iterdir():
                if child.is_file():
                    datafiles[child.name] = child.read_bytes()
                    log.info(f"read {child} into data path as {child.name}")
                else:
                    log.info(f"ignoring {child} in data path")
    return datafiles


def get_deck_configuration() -> DeckConfigurationType:
    """Return the host robot's current deck configuration.

    This is a hacky implementation because the deck configuration is owned by
    robot-server. See `robot_server.deck_configuration.cli`.
    """
    match ARCHITECTURE:
        # These hard-coded paths need to be kept in sync with robot-server's
        # systemd configuration:
        case SystemArchitecture.BUILDROOT:
            robot_server_persistence_dir = "/data/opentrons_robot_server"
        case SystemArchitecture.YOCTO:
            robot_server_persistence_dir = "/var/lib/opentrons-robot-server"
        case SystemArchitecture.HOST:
            # todo(mm, 2024-08-13):
            # It's unclear what should happen if you run an `opentrons.execute` entry
            # point on a non-robot device that doesn't have robot-server. I can't
            # imagine we'd let actual users do this, but it might happen in internal
            # testing. This empty list return value is totally arbitrary and will
            # probably not do the right thing.
            return []

    proc = subprocess.run(
        [
            sys.executable,
            "-m",
            "robot_server.deck_configuration.cli",
            "--persistence-directory",
            robot_server_persistence_dir,
        ],
        check=True,
        capture_output=True,
    )
    deserialized_deck_configuration = deserialize_deck_configuration(proc.stdout)
    if deserialized_deck_configuration is None:
        raise RuntimeError("Error getting the host robot's deck configuration.")

    cutout_fixture_placements, _ = deserialized_deck_configuration

    return [
        (
            cutout_fixture_placement.cutout_id,
            cutout_fixture_placement.cutout_fixture_id,
            cutout_fixture_placement.opentrons_module_serial_number,
        )
        for cutout_fixture_placement in cutout_fixture_placements
    ]


@contextlib.contextmanager
def adapt_protocol_source(protocol: Protocol) -> Generator[ProtocolSource, None, None]:
    """Convert a `Protocol` to a `ProtocolSource`.

    `Protocol` and `ProtocolSource` do basically the same thing. `Protocol` is the traditional
    interface. `ProtocolSource` is the newer, but not necessarily better, interface that's required
    to run stuff through Protocol Engine. Ideally, the two would be unified. Until then, we have
    this shim.

    This is a context manager because it needs to keep some temp files around.
    """
    # ProtocolReader needs to know the filename of the main protocol file so it can infer from its
    # extension whether it's a JSON or Python protocol. But that filename doesn't necessarily exist,
    # like when a user passes a text stream to opentrons.simulate.simulate(). As a hack, work
    # backwards and synthesize a dummy filename with the correct extension.
    if protocol.filename is not None:
        # We were given a filename, so no need to guess.
        #
        # It's not well-defined in our customer-facing interfaces whether the supplied protocol_name
        # should be just the filename part, or a path with separators. In case it contains stuff
        # like "../", sanitize it to just the filename part so we don't save files somewhere bad.
        main_file_name = pathlib.Path(protocol.filename).name
    elif isinstance(protocol, JsonProtocol):
        main_file_name = "protocol.json"
    else:
        main_file_name = "protocol.py"

    with tempfile.TemporaryDirectory() as temporary_directory:
        # FIXME(mm, 2023-06-26): Copying these files is pure overhead, and it introduces encoding
        # hazards. Remove this when we can parse JSONv6+ and PAPIv2.14+ protocols without going
        # through the filesystem. https://opentrons.atlassian.net/browse/RSS-281

        main_file = pathlib.Path(temporary_directory) / main_file_name
        if isinstance(protocol.text, str):
            main_file.write_text(protocol.text, encoding="utf-8")
        else:
            main_file.write_bytes(protocol.text)

        labware_files: List[pathlib.Path] = []
        if isinstance(protocol, PythonProtocol) and protocol.extra_labware is not None:
            for labware_index, labware_definition in enumerate(
                protocol.extra_labware.values()
            ):
                new_labware_file = (
                    pathlib.Path(temporary_directory) / f"{labware_index}.json"
                )
                new_labware_file.write_text(
                    json.dumps(labware_definition), encoding="utf-8"
                )
                labware_files.append(new_labware_file)

        protocol_source = asyncio.run(
            ProtocolReader().read_saved(
                files=[main_file] + labware_files,
                directory=None,
                files_are_prevalidated=False,
            )
        )

        yield protocol_source


class ProtocolEngineExecuteError(Exception):
    def __init__(self, errors: List[ProtocolEngineErrorOccurrence]) -> None:
        """Raised when there was any fatal error running a protocol through Protocol Engine.

        Protocol Engine reports errors as data, not as exceptions.
        But the only way for `opentrons.execute.execute()` and `opentrons.simulate.simulate()`
        to signal problems to their callers is to raise something.
        So we need this class to wrap them.

        Params:
            errors: The errors that Protocol Engine reported.
        """
        # Show the full error details if this is part of a traceback. Don't try to summarize.
        super().__init__(errors)

        self._error_occurrences = errors

    def to_stderr_string(self) -> str:
        """Return a string suitable as the stderr output of the `opentrons_execute` CLI.

        This summarizes from the full error details.
        """
        # It's unclear what exactly we should extract here.
        #
        # First, do we print the first element, or the last, or all of them?
        #
        # Second, do we print the .detail? .errorCode? .errorInfo? .wrappedErrors?
        # By contract, .detail seems like it would be insufficient, but experimentally,
        # it includes a lot, like:
        #
        #     ProtocolEngineError [line 3]: Error 4000 GENERAL_ERROR (ProtocolEngineError):
        #     UnexpectedProtocolError: Labware "fixture_12_trough" not found with version 1
        #     in namespace "fixture".
        return self._error_occurrences[0].detail



"""Command models to execute a Thermocycler profile."""
from __future__ import annotations
from typing import List, Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from opentrons.hardware_control.modules.types import ThermocyclerStep

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


RunProfileCommandType = Literal["thermocycler/runProfile"]


class RunProfileStepParams(BaseModel):
    """Input parameters for an individual Thermocycler profile step."""

    celsius: float = Field(..., description="Target temperature in °C.")
    holdSeconds: float = Field(
        ..., description="Time to hold target temperature at in seconds."
    )


class RunProfileParams(BaseModel):
    """Input parameters to run a Thermocycler profile."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")
    profile: List[RunProfileStepParams] = Field(
        ...,
        description="Array of profile steps with target temperature and temperature hold time.",
    )
    blockMaxVolumeUl: Optional[float] = Field(
        None,
        description="Amount of liquid in uL of the most-full well"
        " in labware loaded onto the thermocycler.",
    )


class RunProfileResult(BaseModel):
    """Result data from running a Thermocycler profile."""


class RunProfileImpl(
    AbstractCommandImpl[RunProfileParams, SuccessData[RunProfileResult]]
):
    """Execution implementation of a Thermocycler's run profile command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: RunProfileParams) -> SuccessData[RunProfileResult]:
        """Run a Thermocycler profile."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        steps = [
            ThermocyclerStep(
                temperature=thermocycler_state.validate_target_block_temperature(
                    profile_step.celsius
                ),
                hold_time_seconds=profile_step.holdSeconds,
            )
            for profile_step in params.profile
        ]

        target_volume: Optional[float]
        if params.blockMaxVolumeUl is not None:
            target_volume = thermocycler_state.validate_max_block_volume(
                params.blockMaxVolumeUl
            )
        else:
            target_volume = None

        if thermocycler_hardware is not None:
            # TODO(jbl 2022-06-27) hardcoded constant 1 for `repetitions` should be
            #  moved from HardwareControlAPI to the Python ProtocolContext
            await thermocycler_hardware.cycle_temperatures(
                steps=steps, repetitions=1, volume=target_volume
            )

        return SuccessData(
            public=RunProfileResult(),
        )


class RunProfile(BaseCommand[RunProfileParams, RunProfileResult, ErrorOccurrence]):
    """A command to execute a Thermocycler profile run."""

    commandType: RunProfileCommandType = "thermocycler/runProfile"
    params: RunProfileParams
    result: Optional[RunProfileResult]

    _ImplementationCls: Type[RunProfileImpl] = RunProfileImpl


class RunProfileCreate(BaseCommandCreate[RunProfileParams]):
    """A request to execute a Thermocycler profile run."""

    commandType: RunProfileCommandType = "thermocycler/runProfile"
    params: RunProfileParams

    _CommandCls: Type[RunProfile] = RunProfile



"""Abstract state store interfaces."""
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from ..actions import Action

StateT = TypeVar("StateT")


class HasState(ABC, Generic[StateT]):
    """Abstract interface for an object that has a state data member."""

    _state: StateT

    @property
    def state(self) -> StateT:
        """State getter."""
        return self._state


class HandlesActions(ABC):
    """Abstract interface for an object that reacts to actions."""

    @abstractmethod
    def handle_action(self, action: Action) -> None:
        """React to a state-change action."""
        ...



"""Opentrons command-line interface.

This CLI application is not yet public-facing.
It is currently used by the Opentrons App to analyze protocols.
"""

import click

from .analyze import analyze


@click.group()
def main() -> None:
    """Welcome to the `opentrons` command-line application.

    This application is not yet public facing.
    Please check back in a future software version.
    """


main.add_command(analyze)



from collections import deque
from typing import Deque

from opentrons.drivers.types import PlateTemperature
from opentrons.hardware_control.modules.types import TemperatureStatus


class PlateTemperatureStatus:

    TEMP_THRESHOLD = 0.3
    """The threshold under which the difference between target and temperature
    is considered `holding`."""

    MIN_SAMPLES_UNDER_THRESHOLD = 10
    """The number of temperature samples needed before determining that the
    temperature is `holding`"""

    def __init__(self) -> None:
        """Construct."""
        self._temp_history: Deque[float] = deque(
            maxlen=self.MIN_SAMPLES_UNDER_THRESHOLD
        )
        self._status = TemperatureStatus.ERROR

    @property
    def status(self) -> TemperatureStatus:
        """Return the current status."""
        return self._status

    def update(self, temperature: PlateTemperature) -> TemperatureStatus:
        """Update the status based on the temperature."""
        # Add to history
        self._temp_history.append(temperature.current)

        if temperature.target is None:
            _status = TemperatureStatus.IDLE
        else:
            diff = temperature.target - temperature.current
            if self._is_holding_at_target(temperature.target, self._temp_history):
                _status = TemperatureStatus.HOLDING
            elif diff < 0:
                _status = TemperatureStatus.COOLING
            else:
                _status = TemperatureStatus.HEATING
        self._status = _status
        return self._status

    @staticmethod
    def _is_holding_at_target(target: float, history: Deque[float]) -> bool:
        """
        Checks block temp history to determine if block temp has stabilized at
        the target temperature. Returns true only if all values in history are
        within threshold range of target temperature.
        """
        if len(history) < PlateTemperatureStatus.MIN_SAMPLES_UNDER_THRESHOLD:
            # Not enough temp history
            return False
        else:
            return all(
                abs(target - t) < PlateTemperatureStatus.TEMP_THRESHOLD for t in history
            )



"""Shared utilities for ot3 hardware control."""
from typing import Dict, Iterable, List, Set, Tuple, TypeVar, cast, Sequence, Optional
from typing_extensions import Literal
from logging import getLogger
from opentrons.config.defaults_ot3 import DEFAULT_CALIBRATION_AXIS_MAX_SPEED
from opentrons.config.types import OT3MotionSettings, OT3CurrentSettings, GantryLoad
from opentrons.hardware_control.types import (
    Axis,
    OT3AxisKind,
    OT3AxisMap,
    CurrentConfig,
    SubSystem,
    OT3Mount,
    InstrumentProbeType,
    PipetteSubType,
    UpdateState,
    UpdateStatus,
    GripperJawState,
)
import numpy as np

from opentrons_hardware.firmware_bindings.constants import (
    NodeId,
    FirmwareTarget,
    PipetteType,
    SensorId,
    PipetteTipActionType,
    USBTarget,
    GripperJawState as FirmwareGripperjawState,
)
from opentrons_hardware.firmware_update.types import FirmwareUpdateStatus, StatusElement
from opentrons_hardware.hardware_control import network
from opentrons_hardware.hardware_control.motion_planning import (
    AxisConstraints,
    SystemConstraints,
    Coordinates,
    Move,
    CoordinateValue,
)
from opentrons_hardware.hardware_control.tool_sensors import (
    InstrumentProbeTarget,
    PipetteProbeTarget,
)
from opentrons_hardware.hardware_control.motion_planning.move_utils import (
    unit_vector_multiplication,
)
from opentrons_hardware.hardware_control.motion import (
    create_step,
    NodeIdMotionValues,
    create_home_step,
    create_backoff_step,
    create_tip_action_backoff_step,
    MoveGroup,
    MoveType,
    MoveStopCondition,
    create_gripper_jaw_step,
    create_tip_action_step,
)
from opentrons_hardware.hardware_control.constants import interrupts_per_sec

GRIPPER_JAW_HOME_TIME: float = 10
GRIPPER_JAW_GRIP_TIME: float = 10

LIMIT_SWITCH_OVERTRAVEL_DISTANCE: float = 1

PipetteAction = Literal["clamp", "home"]

# TODO: These methods exist to defer uses of NodeId to inside
# method bodies, which won't be evaluated until called. This is needed
# because the robot server doesn't have opentrons_ot3_firmware as a dep
# which is where they're defined, and therefore you can't have references
# to NodeId that are interpreted at import time because then the robot
# server tests fail when importing hardware controller. This is obviously
# terrible and needs to be fixed.

SUBSYSTEM_NODEID: Dict[SubSystem, NodeId] = {
    SubSystem.gantry_x: NodeId.gantry_x,
    SubSystem.gantry_y: NodeId.gantry_y,
    SubSystem.head: NodeId.head,
    SubSystem.pipette_left: NodeId.pipette_left,
    SubSystem.pipette_right: NodeId.pipette_right,
    SubSystem.gripper: NodeId.gripper,
    SubSystem.hepa_uv: NodeId.hepa_uv,
}

NODEID_SUBSYSTEM = {node: subsystem for subsystem, node in SUBSYSTEM_NODEID.items()}

SUBSYSTEM_USB: Dict[SubSystem, USBTarget] = {SubSystem.rear_panel: USBTarget.rear_panel}

USB_SUBSYSTEM = {target: subsystem for subsystem, target in SUBSYSTEM_USB.items()}

LOG = getLogger(__name__)


def axis_nodes() -> List["NodeId"]:
    return [
        NodeId.gantry_x,
        NodeId.gantry_y,
        NodeId.head_l,
        NodeId.head_r,
        NodeId.pipette_left,
        NodeId.pipette_right,
        NodeId.gripper_z,
        NodeId.gripper_g,
    ]


def node_axes() -> List[Axis]:
    return Axis.node_axes()


def home_axes() -> List[Axis]:
    return [
        Axis.P_L,
        Axis.P_R,
        Axis.G,
        Axis.Z_L,
        Axis.Z_R,
        Axis.Z_G,
        Axis.X,
        Axis.Y,
    ]


def axis_to_node(axis: Axis) -> "NodeId":
    anm = {
        Axis.X: NodeId.gantry_x,
        Axis.Y: NodeId.gantry_y,
        Axis.Z_L: NodeId.head_l,
        Axis.Z_R: NodeId.head_r,
        Axis.P_L: NodeId.pipette_left,
        Axis.P_R: NodeId.pipette_right,
        Axis.Z_G: NodeId.gripper_z,
        Axis.G: NodeId.gripper_g,
        Axis.Q: NodeId.pipette_left,
    }
    return anm[axis]


def node_to_axis(node: "NodeId") -> Axis:
    nam = {
        NodeId.gantry_x: Axis.X,
        NodeId.gantry_y: Axis.Y,
        NodeId.head_l: Axis.Z_L,
        NodeId.head_r: Axis.Z_R,
        NodeId.pipette_left: Axis.P_L,
        NodeId.pipette_right: Axis.P_R,
        NodeId.gripper_z: Axis.Z_G,
        NodeId.gripper_g: Axis.G,
    }
    return nam[node]


def node_is_axis(node: "NodeId") -> bool:
    try:
        node_to_axis(node)
        return True
    except KeyError:
        return False


def axis_is_node(axis: Axis) -> bool:
    try:
        axis_to_node(axis)
        return True
    except KeyError:
        return False


def sub_system_to_nodeid(sub_sys: SubSystem) -> "NodeId":
    """Convert a sub system to a NodeId."""
    return SUBSYSTEM_NODEID[sub_sys]


def node_id_to_subsystem(node_id: NodeId) -> "SubSystem":
    """Convert a NodeId to a Subsystem"""
    return NODEID_SUBSYSTEM[node_id.application_for()]


def usb_to_subsystem(target: USBTarget) -> SubSystem:
    return USB_SUBSYSTEM[target]


def subsystem_to_usb(subsystem: SubSystem) -> USBTarget:
    return SUBSYSTEM_USB[subsystem]


def target_to_subsystem(target: FirmwareTarget) -> SubSystem:
    if isinstance(target, USBTarget):
        return usb_to_subsystem(target)
    elif isinstance(target, NodeId):
        return node_id_to_subsystem(target)
    else:
        raise KeyError(target)


def subsystem_to_target(subsystem: SubSystem) -> FirmwareTarget:
    try:
        return sub_system_to_nodeid(subsystem)
    except KeyError:
        return subsystem_to_usb(subsystem)


def get_current_settings(
    config: OT3CurrentSettings,
    gantry_load: GantryLoad,
) -> OT3AxisMap[CurrentConfig]:
    conf_by_pip = config.by_gantry_load(gantry_load)
    currents = {}
    for axis_kind in conf_by_pip["hold_current"].keys():
        for axis in Axis.of_kind(axis_kind):
            currents[axis] = CurrentConfig(
                hold_current=conf_by_pip["hold_current"][axis_kind],
                run_current=conf_by_pip["run_current"][axis_kind],
            )
    if gantry_load == GantryLoad.HIGH_THROUGHPUT:
        # In high-throughput configuration, the right mount doesn't do anything: the
        # lead screw nut is disconnected from the carriage, and it just hangs out
        # up at the top of the axis. We should therefore not give it a lot of current.
        # TODO: think of a better way to do this
        lt_config = config.by_gantry_load(GantryLoad.LOW_THROUGHPUT)
        currents[Axis.Z_R] = CurrentConfig(
            hold_current=lt_config["hold_current"][OT3AxisKind.Z],
            # not a typo: keep that current low
            run_current=lt_config["hold_current"][OT3AxisKind.Z],
        )
    return currents


def get_system_constraints(
    config: OT3MotionSettings,
    gantry_load: GantryLoad,
) -> "SystemConstraints[Axis]":
    conf_by_pip = config.by_gantry_load(gantry_load)
    constraints = {}
    axis_kind_list = [
        OT3AxisKind.P,
        OT3AxisKind.X,
        OT3AxisKind.Y,
        OT3AxisKind.Z,
        OT3AxisKind.Z_G,
    ]
    if gantry_load == GantryLoad.HIGH_THROUGHPUT:
        axis_kind_list.append(OT3AxisKind.Q)
    for axis_kind in axis_kind_list:
        for axis in Axis.of_kind(axis_kind):
            constraints[axis] = AxisConstraints.build(
                conf_by_pip["acceleration"][axis_kind],
                conf_by_pip["max_speed_discontinuity"][axis_kind],
                conf_by_pip["direction_change_speed_discontinuity"][axis_kind],
                conf_by_pip["default_max_speed"][axis_kind],
            )
    return constraints


def get_system_constraints_for_calibration(
    config: OT3MotionSettings,
    gantry_load: GantryLoad,
) -> "SystemConstraints[Axis]":
    conf_by_pip = config.by_gantry_load(gantry_load)
    constraints = {}
    for axis_kind in [
        OT3AxisKind.P,
        OT3AxisKind.X,
        OT3AxisKind.Y,
        OT3AxisKind.Z,
        OT3AxisKind.Z_G,
    ]:
        for axis in Axis.of_kind(axis_kind):
            constraints[axis] = AxisConstraints.build(
                conf_by_pip["acceleration"][axis_kind],
                conf_by_pip["max_speed_discontinuity"][axis_kind],
                conf_by_pip["direction_change_speed_discontinuity"][axis_kind],
                DEFAULT_CALIBRATION_AXIS_MAX_SPEED,
            )
    return constraints


def get_system_constraints_for_plunger_acceleration(
    config: OT3MotionSettings,
    gantry_load: GantryLoad,
    mount: OT3Mount,
    acceleration: float,
) -> "SystemConstraints[Axis]":
    old_constraints = config.by_gantry_load(gantry_load)
    new_constraints = {}
    axis_kinds = set([k for _, v in old_constraints.items() for k in v.keys()])
    for axis_kind in axis_kinds:
        for axis in Axis.of_kind(axis_kind):
            if axis == Axis.of_main_tool_actuator(mount):
                _accel = acceleration
            else:
                _accel = old_constraints["acceleration"][axis_kind]
            new_constraints[axis] = AxisConstraints.build(
                _accel,
                old_constraints["max_speed_discontinuity"][axis_kind],
                old_constraints["direction_change_speed_discontinuity"][axis_kind],
                old_constraints["default_max_speed"][axis_kind],
            )
    return new_constraints


def _convert_to_node_id_dict(
    axis_pos: Coordinates[Axis, CoordinateValue],
) -> NodeIdMotionValues:
    target: NodeIdMotionValues = {}
    for axis, pos in axis_pos.items():
        if axis_is_node(axis):
            target[axis_to_node(axis)] = np.float64(pos)
    return target


def replace_head_node(targets: Set[FirmwareTarget]) -> Set[FirmwareTarget]:
    """Replace the head core node with its two sides.

    The node ID for the head central controller is what shows up in a network probe,
    but what we actually send commands to an overwhelming majority of the time is
    the head_l and head_r synthetic node IDs, and those are what we want in the
    network map.
    """
    if NodeId.head in targets:
        targets.remove(NodeId.head)
        targets.add(NodeId.head_r)
        targets.add(NodeId.head_l)
    return targets


def replace_gripper_node(targets: Set[FirmwareTarget]) -> Set[FirmwareTarget]:
    """Replace the gripper core node with its two axes.

    The node ID for the gripper controller is what shows up in a network probe,
    but what we actually send most commands to is the gripper_z and gripper_g
    synthetic nodes, so we should have them in the network map instead.
    """
    if NodeId.gripper in targets:
        targets.remove(NodeId.gripper)
        targets.add(NodeId.gripper_z)
        targets.add(NodeId.gripper_g)
    return targets


def motor_nodes(devices: Set[FirmwareTarget]) -> Set[NodeId]:
    # do the replacement of head and gripper devices
    motor_nodes = replace_gripper_node(devices)
    motor_nodes = replace_head_node(motor_nodes)
    bootloader_nodes = {
        NodeId.pipette_left_bootloader,
        NodeId.pipette_right_bootloader,
        NodeId.gantry_x_bootloader,
        NodeId.gantry_y_bootloader,
        NodeId.head_bootloader,
        NodeId.gripper_bootloader,
    }
    hepa_uv_nodes = {
        NodeId.hepa_uv,
        NodeId.hepa_uv_bootloader,
    }
    # remove any bootloader nodes
    motor_nodes -= bootloader_nodes
    motor_nodes -= hepa_uv_nodes
    # filter out usb nodes
    return {NodeId(target) for target in motor_nodes if target in NodeId}


def create_move_group(
    origin: Coordinates[Axis, CoordinateValue],
    moves: List[Move[Axis]],
    present_nodes: Iterable[NodeId],
    stop_condition: MoveStopCondition = MoveStopCondition.none,
) -> Tuple[MoveGroup, Dict[NodeId, float]]:
    pos = _convert_to_node_id_dict(origin)
    move_group: MoveGroup = []
    for move in moves:
        unit_vector = move.unit_vector
        for block in move.blocks:
            if block.time < (3.0 / interrupts_per_sec):
                LOG.info(
                    f"Skipping move block with time {block.time} (<{3.0/interrupts_per_sec})"
                )
                continue
            distances = unit_vector_multiplication(unit_vector, block.distance)
            node_id_distances = _convert_to_node_id_dict(distances)
            velocities = unit_vector_multiplication(unit_vector, block.initial_speed)
            accelerations = unit_vector_multiplication(unit_vector, block.acceleration)
            step = create_step(
                distance=node_id_distances,
                velocity=_convert_to_node_id_dict(velocities),
                acceleration=_convert_to_node_id_dict(accelerations),
                duration=block.time,
                present_nodes=present_nodes,
                stop_condition=stop_condition,
            )
            for ax in pos.keys():
                pos[ax] += node_id_distances.get(ax, 0)
            move_group.append(step)
    return move_group, {k: float(v) for k, v in pos.items()}


def create_home_groups(
    distance: Dict[Axis, float], velocity: Dict[Axis, float]
) -> List[MoveGroup]:
    node_id_distances = _convert_to_node_id_dict(distance)
    node_id_velocities = _convert_to_node_id_dict(velocity)
    home_group = [
        create_home_step(distance=node_id_distances, velocity=node_id_velocities)
    ]
    # halve the homing speed for backoff
    backoff_velocities = {k: v / 2 for k, v in node_id_velocities.items()}
    backoff_group = [create_backoff_step(backoff_velocities)]
    return [home_group, backoff_group]


def create_tip_action_group(
    moves: Sequence[Move[Axis]],
    present_nodes: Iterable[NodeId],
    action: str,
) -> MoveGroup:
    move_group: MoveGroup = []
    for move in moves:
        unit_vector = move.unit_vector
        for block in move.blocks:
            if block.time < (3.0 / interrupts_per_sec):
                continue
            velocities = unit_vector_multiplication(unit_vector, block.initial_speed)
            accelerations = unit_vector_multiplication(unit_vector, block.acceleration)
            step = create_tip_action_step(
                velocity=_convert_to_node_id_dict(velocities),
                acceleration=_convert_to_node_id_dict(accelerations),
                duration=block.time,
                present_nodes=present_nodes,
                action=PipetteTipActionType[action],
            )
            move_group.append(step)
    return move_group


def create_tip_motor_home_group(
    distance: float,
    velocity: float,
    backoff: Optional[bool] = False,
) -> MoveGroup:
    move_group: MoveGroup = []
    home_step = create_tip_action_step(
        velocity={NodeId.pipette_left: np.float64(-1 * velocity)},
        acceleration={NodeId.pipette_left: np.float64(0)},
        duration=np.float64(distance / velocity),
        present_nodes=[NodeId.pipette_left],
        action=PipetteTipActionType.home,
    )
    move_group.append(home_step)

    if backoff:
        backoff_group = create_tip_action_backoff_step(
            velocity={
                node_id: np.float64(velocity / 2) for node_id in [NodeId.pipette_left]
            }
        )
        move_group.append(backoff_group)
    return move_group


def create_gripper_jaw_grip_group(
    duty_cycle: float,
    stop_condition: MoveStopCondition = MoveStopCondition.none,
    stay_engaged: bool = True,
) -> MoveGroup:
    step = create_gripper_jaw_step(
        duration=np.float64(GRIPPER_JAW_GRIP_TIME),
        duty_cycle=np.float32(round(duty_cycle)),
        stop_condition=stop_condition,
        move_type=MoveType.grip,
        stay_engaged=stay_engaged,
    )
    move_group: MoveGroup = [step]
    return move_group


def create_gripper_jaw_home_group(dc: float) -> MoveGroup:
    step = create_gripper_jaw_step(
        duration=np.float64(GRIPPER_JAW_HOME_TIME),
        duty_cycle=np.float32(dc),
        stop_condition=MoveStopCondition.limit_switch,
        move_type=MoveType.home,
    )
    move_group: MoveGroup = [step]
    return move_group


def create_gripper_jaw_hold_group(encoder_position_um: int) -> MoveGroup:
    step = create_gripper_jaw_step(
        duration=np.float64(GRIPPER_JAW_GRIP_TIME),
        duty_cycle=np.float32(0),
        encoder_position_um=np.int32(encoder_position_um),
        stop_condition=MoveStopCondition.encoder_position,
        move_type=MoveType.linear,
    )
    move_group: MoveGroup = [step]
    return move_group


def moving_pipettes_in_move_group(group: MoveGroup) -> List[NodeId]:
    """Utility function to get which pipette nodes are moving either in z or their plunger."""
    all_nodes = [node for step in group for node, _ in step.items()]
    moving_nodes = moving_axes_in_move_group(group)
    pipettes_moving: List[NodeId] = [
        k for k in moving_nodes if k in [NodeId.pipette_left, NodeId.pipette_right]
    ]
    if NodeId.head_l in moving_nodes and NodeId.pipette_left in all_nodes:
        pipettes_moving.append(NodeId.pipette_left)
    if NodeId.head_r in moving_nodes and NodeId.pipette_right in all_nodes:
        pipettes_moving.append(NodeId.pipette_right)
    return pipettes_moving


def moving_axes_in_move_group(group: MoveGroup) -> Set[NodeId]:
    """Utility function to get only the moving nodes in a move group."""
    ret: Set[NodeId] = set()
    for step in group:
        for node, node_step in step.items():
            if node_step.is_moving_step():
                ret.add(node)
    return ret


AxisMapPayload = TypeVar("AxisMapPayload")


def axis_convert(
    axis_map: Dict["NodeId", AxisMapPayload], default_value: AxisMapPayload
) -> OT3AxisMap[AxisMapPayload]:
    ret: OT3AxisMap[AxisMapPayload] = {k: default_value for k in node_axes()}
    for node, value in axis_map.items():
        if node_is_axis(node):
            ret[node_to_axis(node)] = value
    return ret


_sensor_node_lookup: Dict[OT3Mount, InstrumentProbeTarget] = {
    OT3Mount.LEFT: NodeId.pipette_left,
    OT3Mount.RIGHT: NodeId.pipette_right,
    OT3Mount.GRIPPER: NodeId.gripper,
}

_sensor_node_lookup_pipettes_only: Dict[OT3Mount, PipetteProbeTarget] = {
    OT3Mount.LEFT: NodeId.pipette_left,
    OT3Mount.RIGHT: NodeId.pipette_right,
}


def sensor_node_for_mount(mount: OT3Mount) -> InstrumentProbeTarget:
    return _sensor_node_lookup[mount]


def sensor_node_for_pipette(mount: OT3Mount) -> PipetteProbeTarget:
    return _sensor_node_lookup_pipettes_only[mount]


_instr_sensor_id_lookup: Dict[InstrumentProbeType, SensorId] = {
    InstrumentProbeType.PRIMARY: SensorId.S0,
    InstrumentProbeType.SECONDARY: SensorId.S1,
    InstrumentProbeType.BOTH: SensorId.BOTH,
}


def sensor_id_for_instrument(probe: InstrumentProbeType) -> SensorId:
    return _instr_sensor_id_lookup[probe]


_pipette_channels_to_sensor_map = {
    PipetteType.pipette_single: [SensorId.S0],
    PipetteType.pipette_multi: [SensorId.S0, SensorId.S1],
    PipetteType.pipette_96: [SensorId.S0, SensorId.S1],
}


def map_pipette_type_to_sensor_id(
    available_instruments: List[NodeId],
    device_info: Dict[SubSystem, network.DeviceInfoCache],
) -> Dict[PipetteProbeTarget, List[SensorId]]:
    return_map = {}
    for node in available_instruments:
        node_info = device_info[node_id_to_subsystem(node)]
        return_map[cast(PipetteProbeTarget, node)] = _pipette_channels_to_sensor_map[
            PipetteType(node_info.subidentifier)
        ]
    return return_map


_pipette_subtype_lookup = {
    PipetteSubType.pipette_single: PipetteType.pipette_single,
    PipetteSubType.pipette_multi: PipetteType.pipette_multi,
    PipetteSubType.pipette_96: PipetteType.pipette_96,
}


def pipette_type_for_subtype(pipette_subtype: PipetteSubType) -> PipetteType:
    return _pipette_subtype_lookup[pipette_subtype]


_update_state_lookup = {
    FirmwareUpdateStatus.queued: UpdateState.queued,
    FirmwareUpdateStatus.updating: UpdateState.updating,
    FirmwareUpdateStatus.done: UpdateState.done,
}


def fw_update_state_from_status(state: FirmwareUpdateStatus) -> UpdateState:
    return _update_state_lookup[state]


class UpdateProgress:
    """Class to keep track of Update progress."""

    def __init__(self, targets: Set[FirmwareTarget]):
        self._tracker: Dict[FirmwareTarget, UpdateStatus] = {}
        self._total_progress = 0
        for target in targets:
            subsystem = (
                node_id_to_subsystem(NodeId(target))
                if isinstance(target, NodeId)
                else SubSystem.rear_panel
            )
            self._tracker[target] = UpdateStatus(subsystem, UpdateState.queued, 0)

    @property
    def targets(self) -> Set[FirmwareTarget]:
        """Gets the set of update Targets queued or updating."""
        return set(self._tracker)

    def get_progress(self) -> Set[UpdateStatus]:
        """Gets the update status and total progress"""
        return set(self._tracker.values())

    def update(
        self, target: FirmwareTarget, status_element: StatusElement
    ) -> Set[UpdateStatus]:
        """Update internal states/progress of firmware updates."""
        fw_update_status, progress = status_element
        subsystem = (
            node_id_to_subsystem(NodeId(target))
            if isinstance(target, NodeId)
            else SubSystem.rear_panel
        )
        state = fw_update_state_from_status(fw_update_status)
        progress = int(progress * 100)
        self._tracker[target] = UpdateStatus(subsystem, state, progress)
        return set(self._tracker.values())


_gripper_jaw_state_lookup: Dict[FirmwareGripperjawState, GripperJawState] = {
    FirmwareGripperjawState.unhomed: GripperJawState.UNHOMED,
    FirmwareGripperjawState.force_controlling_home: GripperJawState.HOMED_READY,
    FirmwareGripperjawState.force_controlling: GripperJawState.GRIPPING,
    FirmwareGripperjawState.position_controlling: GripperJawState.HOLDING,
    FirmwareGripperjawState.stopped: GripperJawState.STOPPED,
}


def gripper_jaw_state_from_fw(state: FirmwareGripperjawState) -> GripperJawState:
    return _gripper_jaw_state_lookup[state]



"""An emulation of the Smoothie.

The purpose is to provide a fake backend that responds to the GCODE sent by the
Opentrons smoothie driver.
"""
import logging
import re
from typing import Optional, Dict

from opentrons import _find_smoothie_file
from opentrons.drivers import utils
from opentrons.drivers.smoothie_drivers.constants import GCODE, HOMED_POSITION
from opentrons.hardware_control.emulation.parser import Command, Parser

from .abstract_emulator import AbstractEmulator
from .settings import SmoothieSettings

logger = logging.getLogger(__name__)


class SmoothieEmulator(AbstractEmulator):
    """Smoothie emulator"""

    WRITE_INSTRUMENT_RE = re.compile(r"(?P<mount>[LR])\s*(?P<value>[a-f0-9]+)")
    INSTRUMENT_AND_MODEL_STRING_LENGTH = 64

    _version_string: str
    _pos: Dict[str, float]
    _home_status: Dict[str, bool]
    _speed: float
    _pipette_model: Dict[str, str]
    _pipette_id: Dict[str, str]

    def __init__(self, parser: Parser, settings: SmoothieSettings) -> None:
        """Constructor.

        Args:
            parser: GCODE Parser.
            settings: emulator settings.
        """
        self._parser = parser
        self._settings = settings
        self._gcode_to_function_mapping = {
            GCODE.HOMING_STATUS.value: self._get_homing_status,
            GCODE.CURRENT_POSITION.value: self._get_current_position,
            GCODE.VERSION.value: self._get_version,
            GCODE.READ_INSTRUMENT_ID.value: self._get_pipette_id,
            GCODE.READ_INSTRUMENT_MODEL.value: self._get_pipette_model,
            GCODE.WRITE_INSTRUMENT_ID.value: self._set_pipette_id,
            GCODE.WRITE_INSTRUMENT_MODEL.value: self._set_pipette_model,
            GCODE.MOVE.value: self._move_gantry,
            GCODE.HOME.value: self._home_gantry,
        }
        self.reset()

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = " ".join(r for r in results if r)
        return None if not joined else joined

    def reset(self) -> None:
        _, fw_version = _find_smoothie_file()
        self._version_string = (
            f"Build version: {fw_version}, Build date: CURRENT, "
            f"MCU: NONE, System Clock: NONE"
        )

        self._pos = {"A": 0.0, "B": 0.0, "C": 0.0, "X": 0.0, "Y": 0.0, "Z": 0.0}
        self._home_status: Dict[str, bool] = {
            "X": False,
            "Y": False,
            "Z": False,
            "A": False,
            "B": False,
            "C": False,
        }
        self._speed = 0.0

        self._pipette_model = {
            "L": utils.string_to_hex(
                self._settings.left.model, self.INSTRUMENT_AND_MODEL_STRING_LENGTH
            ),
            "R": utils.string_to_hex(
                self._settings.right.model, self.INSTRUMENT_AND_MODEL_STRING_LENGTH
            ),
        }

        self._pipette_id = {
            "L": utils.string_to_hex(
                self._settings.left.id, self.INSTRUMENT_AND_MODEL_STRING_LENGTH
            ),
            "R": utils.string_to_hex(
                self._settings.right.id, self.INSTRUMENT_AND_MODEL_STRING_LENGTH
            ),
        }

    def get_current_position(self) -> Dict[str, float]:
        return self._pos

    def _get_homing_status(self, command: Command) -> str:
        """Get the current homing status of the emulated gantry"""
        return " ".join(f"{k}:{int(v)}" for k, v in self._home_status.items())

    def _get_current_position(self, command: Command) -> str:
        """Get the current position of the emulated gantry"""
        pos_string = " ".join(f"{k}:{v}" for k, v in self._pos.items())
        return f"{command.gcode}\r\n\r\nok MCS: {pos_string}"

    def _get_version(self, command: Command) -> str:
        """Get the current firmware version"""
        return self._version_string

    def _get_pipette_id(self, command: Command) -> Optional[str]:
        """Get the current id of the specified pipette"""
        pipette_postion = None

        if "L" in command.params:
            pipette_postion = f"L:{self._pipette_id['L']}"
        elif "R" in command.params:
            pipette_postion = f"R:{self._pipette_id['R']}"

        return pipette_postion

    def _get_pipette_model(self, command: Command) -> Optional[str]:
        """Get the current model of the specified pipette"""
        pipette_model = None

        if "L" in command.params:
            pipette_model = f"L:{self._pipette_model['L']}"
        elif "R" in command.params:
            pipette_model = f"R:{self._pipette_model['R']}"

        return pipette_model

    def _set_pipette_id(self, command: Command) -> None:
        """Sets the id for the pipette"""
        self._pipette_id.update(self._mount_strings(command))

    def _set_pipette_model(self, command: Command) -> None:
        """Sets the model for the pipette"""
        self._pipette_model.update(self._mount_strings(command))

    def _move_gantry(self, command: Command) -> None:
        """Moves the gantry to the position provided in the command"""
        for key, value in command.params.items():
            assert isinstance(value, float), f"invalid value '{value}'"
            if "F" == key:
                self._speed = value
            else:
                self._pos[key] = value

    def _home_gantry(self, command: Command) -> None:
        """Returns gantry to home position"""
        for axis in command.params.keys():
            self._pos[axis] = HOMED_POSITION[axis]
            self._home_status[axis] = True

    def _handle(self, command: Command) -> Optional[str]:
        """Handle a command."""
        logger.info(f"Got command {command}")
        func_to_run = self._gcode_to_function_mapping.get(command.gcode)
        return None if func_to_run is None else func_to_run(command)

    @staticmethod
    def _mount_strings(command: Command) -> Dict[str, str]:
        """
        Parse the body of the command for the mount strings.

        Write instrument id and model use the format:
            MOUNT VALUE
        where MOUNT is either L or R and VALUE is a string of ascii codes.

        Args:
            command: Command

        Returns:
            A dict of L and/or R to the string value following it.
        """
        pars = (
            i.groupdict()
            for i in SmoothieEmulator.WRITE_INSTRUMENT_RE.finditer(command.body)
        )
        result = {
            p["mount"]: p["value"]
            + "0"
            * (SmoothieEmulator.INSTRUMENT_AND_MODEL_STRING_LENGTH - len(p["value"]))
            for p in pars
        }

        assert result, f"missing mount values '{command.body}'"
        return result



"""Common APIs for protocol core interfaces."""

# TODO(mc, 2022-08-22): move to __init__ when dependency cycles are resolved
from .instrument import AbstractInstrument
from .labware import AbstractLabware
from .module import (
    AbstractModuleCore,
    AbstractTemperatureModuleCore,
    AbstractMagneticModuleCore,
    AbstractThermocyclerCore,
    AbstractHeaterShakerCore,
    AbstractMagneticBlockCore,
    AbstractAbsorbanceReaderCore,
)
from .protocol import AbstractProtocol
from .well import AbstractWellCore


WellCore = AbstractWellCore
LabwareCore = AbstractLabware[WellCore]
InstrumentCore = AbstractInstrument[WellCore]
ModuleCore = AbstractModuleCore
TemperatureModuleCore = AbstractTemperatureModuleCore
MagneticModuleCore = AbstractMagneticModuleCore
ThermocyclerCore = AbstractThermocyclerCore
HeaterShakerCore = AbstractHeaterShakerCore
MagneticBlockCore = AbstractMagneticBlockCore
AbsorbanceReaderCore = AbstractAbsorbanceReaderCore
ProtocolCore = AbstractProtocol[InstrumentCore, LabwareCore, ModuleCore]
